
import math
from random import randint, choice
import textdungeon.color as color
import textdungeon.entities as entities
from textdungeon.extra import gather_input, slowprint, add_message

c = color

player = entities.player

class Item:
    name = "item"
    value = 10
    enchantValue = 0.4 # +20% of price per level of enchantment
    maxUses = 1

    enchantable = False
    enchantment = 0
    
    usePrompt = None # if this is None, then the inventory won't provide use as an options
    
    canDig = False # determines if an item can break walls

    def __init__(self):
        self.value = int(self.value * randint(9, 11) / 10) # 90% to 110% of value
        self.uses = self.maxUses

    def degrade(self):
        # if INT is less than 0, there's a 10 * INT % chance that item degrades twice
        if player.intelligence < 0:
            if -randint(0, 99) > player.intelligence * 10:
                self.uses -= 1
            self.uses -= 1
        # for every level of INT, there is a 10% that the item doesn't degrade
        else:
            if randint(0, 99) < player.intelligence * 10:
                return False
            else:
                self.uses -= 1

        # destroys item if uses are at 0
        if self.uses == 0:
            # if it is reusable it says when it breaks
            if self.maxUses > 1:
                add_message(c.red(self.name + " has broken"))

            self.discard()
            for key in player.gear.keys():
                if player.gear[key] == self:
                    player.gear[key] = None
                    self.unequip()
            
            player.inventory.remove(self)
        return True

    def status(self):
        return f"({self.uses} uses)"

    def get_name(self):
        message = ""

        if self.status() != "":
            message += self.status() + " "

        message += self.name

        if self.enchantment > 0:
            message += c.green(f" (+{self.enchantment})")
        elif self.enchantment < 0:
            message += c.red(f" (-{-self.enchantment})")

        return message
        
    def inspect(self):
        print(f"This is a {self.name} with {self.status()} uses")
        
    def attack(self, enemies):
    # performs the items use in combat
        print(f"{self.name} has no use here")
        return False

    def consume(self, floor):
    # performs the items use when there is no special prompt
        print(f"{self.name} has no use here")
        return False

    def unlock(self, lockType):
    # checks if the item can open a lockwall
        print(f"{self.name} isn't a key")
        return False

    def dig(self):
    # checks if the item can dig through a wall
        print(f"{self.name} can't dig")
        return False

    def pickup(self):
    # runs when an item is added to inventory
        return False

    def discard(self):
    # runs when an item is dropped or destroyed, reverses pickup
        return False

    def recharge(self):
    # runs for every item when player descends a floor
        return False
    
    def get_price(self):
        return int((self.value + (self.value * self.enchantValue * self.enchantment)) * (self.uses / self.maxUses / 2 + 0.5))

class Weapon(Item):
    enchantable = True
    enchantValue = 0.2
    
    def __init__(self, name, level):
        if type(self) in [Sword, Mace, Spear, Dagger]:
            material = ["bronze", "iron", "steel", "mithril"][level]
    
            self.name = material + " " + name
            self.value = 50 * level
            self.maxUses = 15 + (10 * level)
    
            self.damage = 4 + level

        super().__init__()

    def degrade(self):
        # if INT is less than 0, there's a 10 * INT % chance that item degrades twice
        if player.intelligence < 0:
            if -randint(0, 99) > player.intelligence * 10:
                self.uses -= 1
            self.uses -= 1
        # for every level of INT, there is a 10% that the item doesn't degrade
        else:
            if randint(0, 99) < player.intelligence * 10:
                return False
            else:
                self.uses -= 1

        # destroys item if uses are at 0
        if self.uses == 0:
            # if it is reusable it says when it breaks
            if self.maxUses > 1:
                add_message(c.red(self.name + " has broken, making it much weaker"))
                self.uses -= 1 # prevents intelligence making this message appear again, doesn't break though
        return True
    
    def status(self):
        suffix = ""
        if self.uses <= 0:
            suffix = "broken"
        elif self.uses <= self.maxUses / 3:
            suffix = "damaged"
        elif self.uses <= self.maxUses * 2 / 3:
            suffix = "worn"

        return f"{suffix}"

    def attack(self, enemies):
        damageDealt = self.damage - 2 * int(self.uses <= 0) + self.enchantment

        if len(enemies) == 1:
            self.target = enemies[0]
        else:
            options = [] # gets a list of enemy names
            for enemy in enemies:
                options.append(enemy.name)
    
            # gets player input
            self.target = enemies[gather_input("Who do you attack?", options)]

        self.degrade()
        
        return damageDealt

class Sword(Weapon):
# does damage to target and can inflict bleeding
# lvl 0 = bronze, lvl 1 = iron, lvl 2 = steel, lvl 3 = mithril
    def __init__(self, level = 0):
        super().__init__("sword", level)

        if type(self) == Sword:
            self.bleedDuration = 4
            self.bleedChance = 2 # bleedChance is _ in 6
    
            if level >= 1:
                self.bleedChance += 1
            if level >= 2:
                self.bleedDuration += 1
            if level >= 3:
                self.bleedChance += 1
                self.bleedDuration += 1

    def inspect(self):
        gcd = math.gcd(self.bleedChance, 6)
        print(f"Does {c.compare(self.damage + self.enchantment, self.damage)} damage, with a {self.bleedChance // gcd} in {6 // gcd} chance to inflict bleeding for {self.bleedDuration} turns.")
        if self.uses < 0:
            print("Because it's broken it does less damage and cannot inflict bleeding.")
    
    def attack(self, enemies):
        damageDealt = super().attack(enemies)

        if self.target.dodge(player):
            print(f"{self.target.name} dodges your attack!")
            return True
            
        damageDealt = self.target.hurt(player, damageDealt)

        if randint(0, 5) < self.bleedChance and self.uses > 0:
            self.target.affect(entities.Bleeding(), self.bleedDuration)
            print(f"You attack {self.target.name} for {c.red(damageDealt)} damage, leaving them {c.effect(entities.Bleeding)}.")
        else:
            print(f"You attack {self.target.name} for {c.red(damageDealt)} damage.")

        return True

class CursedSword(Sword):
# same as sword but extra damage and burning against undead
    name = "Cursed Blade"
    value = 100
    maxUses = 16

    enchantment = -1
    enchantable = False

    damage = 5
    bleedChance = 3
    bleedDuration = 7

    usePrompt = "sacrifice health"
    
    def inspect(self):
        self.enchantment *= -1
        super().inspect()
        self.enchantment *= -1
        print(f"Becomes stronger when cursed {c.red('(-)')}, but weaker when blessed {c.green('(+)')}.")
        print("You can permanently sacrifice health to increase the swords strength. ")
        print(f"You currently have {c.health_status(player.health, player.maxHealth)} health.")

    def consume(self, floor):
        self.uses = self.maxUses
        healthLost = 1
        # becomes more expensive depending the level of curse
        if self.enchantment < 0:
            healthLost -= self.enchantment
            
        self.enchantment -= 1
        self.bleedDuration += 1
        player.maxHealth -= healthLost
        player.health -= healthLost

        player.affect(entities.Bleeding(), healthLost)

        print(f"Your sacrifice of {c.red(healthLost)} health makes the blade stronger.")
        return True
    
    def attack(self, enemies):
        self.enchantment *= -1
        super().attack(enemies)
        self.enchantment *= -1

        return True

class Spear(Weapon):
# does damage to target and has some armor piercing
# lvl 0 = bronze, lvl 1 = iron, lvl 2 = steel, lvl 3 = mithril
    def __init__(self, level = 0):
        super().__init__("spear", level)

        if type(self) == Spear:
            self.armorPiercing = (level + 3) // 2 # level/AP : 0/1, 1/2, 2/2, 3/3

    def inspect(self):
        print(f"Does {c.compare(self.damage + self.enchantment, self.damage)} damage and pierces {self.armorPiercing} points of armor class.")
        if self.uses < 0:
            print("Because it's broken it does less damage and cannot pierce armor.")
    
    def attack(self, enemies):
        damageDealt = super().attack(enemies)

        if self.target.dodge(player):
            print(f"{self.target.name} dodges your attack!")
            return True
            
        damageDealt = self.target.hurt(player, damageDealt, self.armorPiercing)

        print(f"You attack {self.target.name} for {c.red(damageDealt)} damage.")
        return True

class EnchantedSpear(Spear):
# same as spear but has high piercing and ignores dodge
    name = "Enchanted Spear"
    value = 100
    maxUses = 25

    damage = 6
    armorPiercing = 10

    def inspect(self):
        super().inspect()
        print("This spear cannot be dodged.")

    def attack(self, enemies):
        dodgeChances = [] # saves and removes enemy dodge chance
        for enemy in enemies:
            dodgeChances.append(enemy.dodgeChance)
            enemy.dodgeChance = 0

        super().attack(enemies)

        for i in range(len(enemies)): # restores enemy dodge chance
            enemies[i].dodgeChance = dodgeChances[i]

        return True

class Mace(Weapon):
# does damage to target and can stun
# lvl 0 = bronze, lvl 1 = iron, lvl 2 = steel, lvl 3 = mithril
    def __init__(self, level = 0):
        super().__init__("mace", level)
        
        if type(self) == Mace:
            self.stunChance = (level + 3) // 2 # _ in 12, level/stunChance : 0/1, 1/2, 2/2, 3/3

    def inspect(self):
        gcd = math.gcd(self.stunChance, 12)
        print(f"Does {c.compare(self.damage + self.enchantment, self.damage)} damage, with a {self.stunChance // gcd} in {12 // gcd} chance to stun enemies.")
        if self.uses < 0:
            print("Because it's broken it does less damage and cannot stun.")
    
    def attack(self, enemies):
        damageDealt = super().attack(enemies)

        if self.target.dodge(player):
            print(f"{self.target.name} dodges your attack!")
            return True
            
        damageDealt = self.target.hurt(player, damageDealt)

        if randint(0, 11) < self.stunChance and self.uses > 0:
            self.target.stunned = True
            print(f"You attack {self.target.name} for {c.red(damageDealt)} damage, leaving them {c.red('stunned')}.")
            return True

        print(f"You attack {self.target.name} for {c.red(damageDealt)} damage.")
        return True

class FlamingMace(Mace):
# same as mace can set enemies on fire
    name = "Flaming Mace"
    value = 100
    maxUses = 20

    damage = 5
    stunChance = 2

    def inspect(self):
        super().inspect()
        print("It sets enemies on fire, dealing damage over time and lowering their armor class (AC).")

    def attack(self, enemies):
        super().attack(enemies)

        self.target.affect(entities.OnFire(), randint(2, 3))

        return True
        
class Dagger(Weapon):
# does damage to target but uses DEX not STR, strong vs surprised enemies
# lvl 0 = bronze, lvl 1 = iron, lvl 2 = steel, lvl 3 = mithril
    def __init__(self, level = 0):
        super().__init__("dagger", level)

        if type(self) == Dagger:
            self.sneakBonus = level + 2

    def inspect(self):
        print(f"Does {c.compare(self.damage + self.enchantment, self.damage)} damage, and {self.sneakBonus} extra damage towards surprised enemies.")
        print("Does not gain any bonuses from strength (STR).")
        if self.uses < 0:
            print("Because it's broken it does less damage and doesn't do extra damage.")

    def attack(self, enemies):
        damageDealt = super().attack(enemies)

        if self.target.dodge(player):
            print(f"{self.target.name} dodges your attack!")
            return True

        if self.target.has_effect(entities.Surprised):
            damageDealt += self.sneakBonus
            print("Sneak attack!")

        damageDealt = self.target.hurt(player, damageDealt, 0, 0)

        print(f"You attack {self.target.name} for {c.red(damageDealt)} damage.")
        return True

class EbonyDagger(Dagger):
# same as a dagger, but gain max health per kill
    name = "Ebony Dagger"
    value = 100
    maxUses = 50

    damage = 5
    sneakBonus = 4

    def inspect(self):
        super().inspect()
        print(f"This durable weapon heals you whenever you get a kill.")

    def attack(self, enemies):
        super().attack(enemies)

        # applies ebony dagger's effect
        if self.target.health <= 0:
            healing = player.heal(randint(5, 7))
            print("You absorb " + c.green(healing) + " health from " + self.target.name + ".")
        
        return True

class Wand(Item):
# recharges some uses every floor
    name = "wand"
    value = 50
    maxUses = 3

    enchantable = True

    reqInt = 1 # level of INT required to use the wand

    def status(self):
        return f"({self.uses + self.enchantment}/{self.maxUses + self.enchantment} charges)"

    def degrade(self):
        # if INT is less than 0, there's a 7.5 * INT % chance that item degrades twice
        if player.intelligence < 0:
            if -randint(0, 99) > player.intelligence * 10:
                self.uses -= 1
            self.uses -= 1
        # for every level of INT, there is a 7.5% that the item doesn't degrade
        else:
            if randint(0, 99) < player.intelligence * 10:
                print("Your wand gets a free charge because of your INT.")
                return False
            else:
                self.uses -= 1
                
        return True

    def recharge(self):
        self.uses += player.recharge
        
        if self.uses > self.maxUses:
            self.uses = self.maxUses

    def attack(self, enemies):
    # chooses a target
        if player.intelligence < self.reqInt:
            print(f"This wand requires {self.reqInt} intelligence (INT) to use!")
            return False
        
        if self.uses + self.enchantment > 0:
            target = enemies[0]
            
            if len(enemies) > 1:
                options = [] # gets a list of enemy names
                for enemy in enemies:
                    options.append(enemy.name)
        
                # gets player input
                target = enemies[gather_input("Who do you attack?", options)]
    
            self.degrade()
    
            return self.cast(target)
        else:
            print("That wand doesn't have enough charges!")
            return False

    def cast(self, target):
        return True

class HarmWand(Wand):
# (req 2 INT, 4 charges) does 5 + x direct damage, inflicts bleeding for x turns
    name = "wand of harming"
    value = 70
    maxUses = 4

    reqInt = 2

    def inspect(self):
        print(f"Does {5 + self.enchantment} direct damage to target.")
        if self.enchantment > 0:
            print(f"Inflicts target with bleeding for {self.enchantment} turns.")
        print("Requires at least 2 intelligence (INT).")

    def cast(self, target):
        target.health -= 5 + self.enchantment

        if self.enchantment > 0:
            target.affect(entities.Bleeding(), self.enchantment)
            print(f"Your wand of harm does {c.red(5 + self.enchantment)} damage to {target.name}, leaving them {c.effect(entities.Bleeding)}.")
            return True

        print(f"Your wand of harm does {c.red(5 + self.enchantment)} damage to {target.name}.")
        return True

class PoisonWand(Wand):
# (req 1 INT, 3 charges) inflicts poisoned for 5 + x turns, does x direct damage
    name = "wand of poison"
    value = 50
    maxUses = 3

    reqInt = 1

    def inspect(self):
        print(f"Inflicts poison to target for {7 + self.enchantment} turns.")
        if self.enchantment > 0:
            print(f"Does {self.enchantment} direct damage.")
        print("Requires at least 1 intelligence (INT).")

    def cast(self, target):
        target.affect(entities.Poisoned(), 7 + self.enchantment)

        if self.enchantment > 0:
            target.health -= self.enchantment
            print(f"Your wand of poison does {c.red(self.enchantment)} damage to {target.name}, and inflicts {c.effect(entities.Poisoned)}.")
            return True

        print(f"Your wand of poison inflicts {target.name} with {c.effect(entities.Poisoned)}.")
        return True

class LightningWand(Wand):
# (req 3 INT, 2 charges) electrocutes target for 5 + x turns
    name = "wand of lightning"
    value = 75
    maxUses = 2

    reqInt = 3

    def inspect(self):
        print(f"Electrocutes the target for {5 + self.enchantment} turns.")
        print(f"Electrocuted targets become stunned every other turn")
        print("Requires at least 3 intelligence (INT).")

    def cast(self, target):
        target.affect(entities.Electrocuted(), 5 + self.enchantment)
        print(f"Your wand of lightning inflicts {target.name} with {c.effect(entities.Electrocuted)}.")
        return True

class TeleportWand(Wand):
# (req 2 INT, 2 charges) teleports target to a random room
    name = "wand of teleportation"
    value = 65
    maxUses = 2

    reqInt = 2

    def inspect(self):
        print("Teleports the target to a random room.")
        print("Requires at least 2 intelligence (INT).")

    def cast(self, target):
        if issubclass(type(target), entities.Boss):
            print(f"{target.name} is too powerful for that spell!")
            return False
            
        currentRoom = player.currentFloor.get_room()

        roomList = []
        for row in player.currentFloor.layout:
            for room in row:
                if not room.blocked and room != currentRoom:
                    roomList.append(room)

        choice(roomList).threats.append(target)
        currentRoom.threats.remove(target)
        
        print(f"{target.name} disappears!")
        target.heal(6) # heals the target a bit
        return True
    
class Equipable(Item):
    usePrompt = "equip"
    gearType = None # replace with a string such as "armor", "ring" or "back"
    
    def status(self):
        suffix = ""
        if self.uses <= self.maxUses / 3:
            suffix = "damaged"
        elif self.uses <= self.maxUses * 2 / 3:
            suffix = "worn"

        return f"{suffix}"

    def get_name(self):
        if self == player.gear[self.gearType]:
            return super().get_name() + " (equipped)"
            
        return super().get_name()

    def put_on(self):
        if player.gear[self.gearType] != None:
            #player.inventory.append(player.armor)
            player.gear[self.gearType].unequip()
        player.gear[self.gearType] = self
        #player.inventory.remove(self)

    def consume(self, floor):
        self.equip()
        self.put_on()

    def attack(self, enemies):
        return self.consume(None)

    def unequip(self):
        return False
    
class Armor(Equipable):
# gives defense to player but lowers DEX
# lvl 0 = bronze, lvl 1 = iron, lvl 2 = steel, lvl 3 = mithril
    enchantable = True
    enchantValue = 0.25

    gearType = "armor"
    
    def __init__(self, level):
        material = ["bronze", "iron", "steel", "mithril"][level]

        self.name = material + " armor"
        self.value = 5 + (55 * level)
        self.maxUses = 20 + (12 * level)

        self.armorClass = level + 2
        self.dexLoss = 1

        if material in ["iron", "steel"]:
            self.dexLoss += 1

        super().__init__()

    def inspect(self):
        print(f"When equipped it gives you {c.compare(self.armorClass + self.enchantment, self.armorClass)} armor class but lowers your dexterity by {self.dexLoss}.")
        print("Worn as armor.")

    def equip(self):
        player.armorClass += self.armorClass + self.enchantment
        player.update_dex(-self.dexLoss)

    def unequip(self):
        # this is where it unequips
        player.armorClass -= self.armorClass + self.enchantment
        player.update_dex(self.dexLoss)

class LeatherArmor(Equipable):
# gives less defense than any other armor, but has no DEX penalty
    name = "leather armor"
    value = 35
    maxUses = 25
    enchantable = True
    enchantValue = 0.4

    gearType = "armor"

    def inspect(self):
        print(f"When equipped it gives you {c.compare(1 + self.enchantment, 1)} armor class.")
        print("Worn as armor.")

    def equip(self):
        player.armorClass += 1 + self.enchantment

    def unequip(self):
        player.armorClass -= 1 + self.enchantment

class MagicRobe(Equipable):
# provides 0 base armor, but increases wand recharge
    name = "magic robe"
    value = 40
    enchantable = True
    enchantValue = 0.8
    maxUses = 20

    gearType = "armor"

    def inspect(self):
        print("When equipped it recharges your wands an additional time each floor.")
        if self.enchantment != 0:
            print(f"It also provides {c.compare(self.enchantment, 0)} armor class.")
        print("Worn as armor.")

    def equip(self):
        # applies stats
        player.armorClass += self.enchantment
        player.recharge += 1

    def unequip(self):
        player.armorClass -= self.enchantment
        player.recharge -= 1

class Cloak(Equipable):
# provides 0 base armor, but +1 stealth
    name = "cloak"
    value = 50
    enchantable = True
    enchantValue = 0.6
    maxUses = 25

    gearType = "back"

    def inspect(self):
        print(f"When equipped it increases your stealth by 1.")
        if self.enchantment != 0:
            print(f"It also provides {c.compare(self.enchantment, 0)} armor class.")
        print("Worn as a back item.")

    def equip(self):
        # applies stats
        player.armorClass += self.enchantment
        player.stealth += 1

    def unequip(self):
        player.armorClass -= self.enchantment
        player.stealth -= 1

class ShadowCloak(Equipable):
# same as cloak, but enchantments affect stealth, not armor
# starts enchanted
    name = "Cloak of Shadows"
    value = 100
    maxUses = 30
    
    enchantable = True
    gearType = "back"

    def inspect(self):
        enchantment = self.enchantment
        if enchantment < 0: # ensures that -1 enchantment means -1 stealth
            enchantment -= 1
            
        print(f"When equipped it increases your dexterity (DEX) by {c.compare(1 + enchantment, 1)}.")
        print("Worn as a back item.")

    def equip(self):
        enchantment = self.enchantment
        if enchantment < 0: # ensures that -1 enchantment means -1 dex
            enchantment -= 1

        # applies stats
        player.update_dex(enchantment + 1)

    def unequip(self):
        enchantment = self.enchantment
        if enchantment < 0: # ensures that -1 enchantment means -1 dex
            enchantment -= 1
        
        player.update_dex(-enchantment - 1)
        
class InfernoRing(Equipable):
# increases strength, but burns you when attacked
    name = "Ring of Rage"
    value = 100

    enchantable = True
    gearType = "ring"

    def inspect(self):
        print(f"When equipped, increases your damage with weapons (excluding daggers) by {c.compare(0.75 + 0.75 * self.enchantment, 0.75)}.")
        print("Worn as a ring.")

    def equip(self):
        player.bonusDamage += (self.enchantment + 1) * 0.75

    def unequip(self):
        player.bonusDamage -= (self.enchantment + 1) * 0.75

class IllusionRing(Equipable):
# increases dodged, when you dodge the attacker is stunned
    name = "Ring of Illusion"
    value = 100

    enchantable = True
    gearType = "ring"

    def inspect(self):
        enchantment = self.enchantment
        if self.enchantment < 0: # negative enchantment is strong enough to reverse the effect
            enchantment -= 1

        print("Whenever you dodge an attack, the attacker gets stunned.")
        print(f"Increases your chance to dodge by {c.compare(5 + 5 * enchantment, 5)}%.")
        print("Worn as a ring.")

    def equip(self):
        enchantment = self.enchantment
        if self.enchantment < 0: # negative enchantment is strong enough to reverse the effect
            enchantment -= 1

        player.illusionRing = True
        player.dodgeChance += 5 + 5 * enchantment

    def unequip(self):
        enchantment = self.enchantment
        if self.enchantment < 0: # negative enchantment is strong enough to reverse the effect
            enchantment -= 1

        player.illusionRing = False
        player.dodgeChance -= 5 + 5 * enchantment

class ArtifactRing(Equipable):
# can accept up to two other rings, then provides the bonuses of each of them
# any enchantment levels are split between the two rings
    name = "Artifact Ring"
    value = 100

    usePrompt = "activate"

    ring1 = None
    ring2 = None

    enchantable = True
    gearType = "ring"

    def inspect(self):
        if self.ring2 == None:
            print("Can absorb the abilities of two other rings.")
        else:
            self.ring2.enchantment = self.enchantment // 2
            self.ring2.inspect()

        if self.ring1 != None:
            self.ring1.enchantment = (self.enchantment + 1) // 2
            self.ring1.inspect()

    def consume(self, floor):
        if self.usePrompt == "activate":
            options = ["cancel"]
            rings = []
            for item in player.inventory:
                if issubclass(type(item), Ring) and item != self:
                    options.append(item.get_name())
                    rings.append(item)

            if len(options) == 1:
                print(c.red("You don't have any rings to insert into Artifact Ring."))
            else:
                playerInput = gather_input("Select a ring to insert.", options, True) - 1
                if playerInput > -1:
                    ring = rings[playerInput]
                    if self.ring1 == None:
                        self.ring1 = ring
                    else:
                        self.ring2 = ring
                        self.usePrompt = "equip"
                    player.inventory.remove(ring)
                    if ring == player.ring:
                        ring.unequip()
                        player.ring = None
                    self.enchantment += ring.enchantment
                    print("Artifact Ring has absorbed the abilities of " + ring.get_name() + ".")
        else:
            super().consume(floor)

    def equip(self):
        self.ring1.enchantment = (self.enchantment + 1) // 2
        self.ring1.equip()
        self.ring2.enchantment = self.enchantment // 2
        self.ring2.equip()

    def unequip(self):
        self.ring1.unequip()
        self.ring2.unequip()

class Ring(Equipable):
# boosts one stat by 1 level
# 0 = stealth, 1 = dodge, 2 = health, 3 = resistance, 4 = awareness
    enchantable = True
    gearType = "ring"

    def __init__(self, ID = -1):
        self.value = 50
        self.maxUses = 1
        self.enchantValue = 0.9
        # decides what stat is boosted
        self.statID = ID
        if ID < 0:
            self.statID = randint(0, 5)
            
        self.stat = ["stealth", "dodge", "health", "resistance", "awareness", "critChance"][self.statID]
        self.name = "ring of " + ["shadows", "evasion", "resilience", "immunity", "vision", "power"][self.statID]

        super().__init__()

    def inspect(self):
        enchantment = self.enchantment
        if self.enchantment < 0: # negative enchantment is strong enough to reverse the effect
            enchantment -= 1

        print([
            f"Increases your stealth by {c.compare(1 + enchantment, 1)} level(s).",
            f"Increases your chance to dodge by {c.compare(5 + 5 * enchantment, 5)}%.",
            f"Increases your health by {c.compare(2 + 2 * enchantment, 2)}.",
            f"Increases your resistance to disease and injury by {c.compare(1 + enchantment, 1)} level(s).",
            f"Increases your awareness of nearby threats by {c.compare(1 + enchantment, 1)} level(s).",
            f"Increases your chance to get a critical hit by {c.compare(10 + 10 * enchantment, 10)}%."
        ][self.statID])
        print("Worn as a ring.")

    def consume(self, floor):
        self.equip()
        self.put_on()

    def equip(self):
        enchantment = self.enchantment
        if self.enchantment < 0: # negative enchantment is strong enough to reverse the effect
            enchantment -= 1
        
        if self.stat == "stealth":
            player.stealth += 1 + enchantment
            
        elif self.stat == "dodge":
            player.dodgeChance += 5 + 5 * enchantment
            
        elif self.stat == "health":
            player.maxHealth += 2 + 2 * enchantment
            player.health += 2 + 2 * enchantment
            
        elif self.stat == "resistance":
            player.resistance += 1 + enchantment

        elif self.stat == "awareness":
            player.awareness += 1 + enchantment

        elif self.stat == "critChance":
            player.critChance += 10 + 10 * enchantment

    def unequip(self):
        enchantment = self.enchantment
        if self.enchantment < 0: # negative enchantment is strong enough to reverse the effect
            enchantment -= 1
            
        if self.stat == "stealth":
            player.stealth -= 1 + enchantment
            
        elif self.stat == "dodge":
            player.dodgeChance -= 5 + 5 * enchantment
            
        elif self.stat == "health":
            player.maxHealth -= 2 + 2 * enchantment
            player.health -= 2 + 2 * enchantment
            
        elif self.stat == "resistance":
            player.resistance -= 1 + enchantment

        elif self.stat == "awareness":
            player.awareness -= 1 + enchantment

        elif self.stat == "critChance":
            player.critChance -= 10 + 10 * enchantment

class Medicine(Item): 
    usePrompt = "use"

    enchantable = True

    healing = 0
    effectApplied = None
    effectDuration = 0
    effectsCured = []

    def attack(self, enemies):
    # attack does the same thing as consume
        return self.consume(None)

    def consume(self, floor):
        self.degrade()

        # cures effects
        removedEffectsIndexes = []
        for i in range(len(player.effects)):
            if type(player.effects[i]) in self.effectsCured:
                removedEffectsIndexes.append(i)

        removedEffects = []
        removedEffectsIndexes.reverse()
        for i in removedEffectsIndexes:
            removedEffects.append(player.effects[i].name)
            player.effects[i].reverse()
            player.effects.pop(i)

        # heals and apples effects
        healingDone = player.heal(self.healing + self.enchantment + randint(-1, 1))

        if self.effectApplied != None:
            player.affect(self.effectApplied(), self.effectDuration + self.enchantment)

        # prints out a message based on healing and removed effects
        message = f"The {self.name} restores {healingDone} health"
        if removedEffects != []:
            message += " and cures "
            if len(removedEffects) == 1:
                message += removedEffects[0]
            else:
                for i in range(len(removedEffects)):
                    if i == len(removedEffects) - 1:
                        message += ", and "
                    elif i > 0:
                        message += ", "

                    message += removedEffects[i]

        print(message + ".")

        return True

class HealingVial(Medicine):
# cures most effects, heals all health
    name = "vial of healing"
    value = 70
    maxUses = 1
    usePrompt = "drink"

    healing = 100
    effectsCured = [entities.RatDisease, entities.BrokenBones, entities.Bleeding, entities.Burned, entities.Poisoned]

    def status(self):
        return ""

    def inspect(self):
        print("Heals all of your health and cures most effects.")

    def degrade(self):
        player.inventory.remove(self)

class Bandage(Medicine):
# cures bleeding, heals some health, and applies regeneration
    name = "bandage"
    value = 40
    maxUses = 3

    healing = 6
    effectApplied = entities.Regeneration
    effectDuration = 4
    effectsCured = [entities.Bleeding]

    def status(self):
        message = ""
        if self.uses == 1:
            message = "(damaged)"
        elif self.uses == 2:
            message = "(worn)"

        return message

    def inspect(self):
        print(f"Heals around {c.compare(10 + 2 * self.enchantment, 10)} health over time.")
        print(f"Has multiple uses and cures bleeding.")
        if self.enchantment > 0:
            print(c.green("This item does extra healing."))
        elif self.enchantment < 0:
            print(c.red("This item does less healing."))
        
        print(f"\nYou currently have {c.health_status(player.health, player.maxHealth)} health.")

    def get_name(self):
        message = self.name

        if self.status() != "":
            message += " " + self.status()

        if self.enchantment > 0:
            message += c.green(f" (+{self.enchantment})")
        elif self.enchantment < 0:
            message += c.red(f" (-{-self.enchantment})")

        return message

# see gen_enemy() in entities.py for explanation
class Rations(Medicine):
# heals a lot of health but can't be used in combat
    name = "rations"
    value = 25
    maxUses = 1

    usePrompt = "consume"

    healing = 7
    effectApplied = entities.WellFed
    effectDuration = 4

    def status(self):
        return ""
        
    def inspect(self):
        print(f"Heals around {c.compare(15 + 3 * self.enchantment, 15)} health over time.")
        print("Cannot be eaten during combat.")
        print(f"\nYou currently have {c.health_status(player.health, player.maxHealth)} health.")
        if self.enchantment > 0:
            print(c.green("This item does extra healing."))
        elif self.enchantment < 0:
            print(c.red("This item does less healing."))

    def attack(self, enemies):
        print("You don't have enough time to eat!")
        return False

    def degrade(self):
        player.inventory.remove(self)

class Scroll(Item):
# one use item that is more powerful with higher intelligence
    maxUses = 1
    
    enchantable = True
    
    usePrompt = "read"

    def degrade(self):
        player.inventory.remove(self)
        return True

    def status(self):
        return ""

    def consume(self, floor):
        if self.enchantment < 0:
            print(c.red(f"Reading the cursed scroll drains {self.enchantment} health from you!"))
            player.health += self.enchantment
        return True

    def inspect(self):
        if self.enchantment > 0:
            print(c.green(f"This scroll is more powerful, it's effects will be boosted by {self.enchantment} INT."))
        elif self.enchantment < 0:
            print(c.red("This scroll may have undesirable effects."))

class ScrollCleanse(Scroll):
# reverses curses from all items, has an INT in 4 chance to enchant
    name = "scroll of cleansing"
    value = 35

    def inspect(self):
        print("Uncurses every cursed item in this floor and your inventory.")
        print("With higher levels of intelligence (INT) some items may also become blessed.")
        super().inspect()
    
    def consume(self, floor):
        power = player.intelligence + self.enchantment # intelligence boosts effectiveness

        # finds all items in the floor
        allItems = []
        allItems.extend(player.inventory)
        for row in player.currentFloor.layout:
            for room in row:
                for item in room.loot:
                    allItems.append(item)

        # uncurses and upgrades them
        upgradedItemsCount = 0
        for item in allItems:
            if type(item) != CursedSword:
                if item in player.gear.values():
                    item.unequip()

                if item.enchantment < 0:
                    item.enchantment = 0

                if randint(0, 8) < power and item.enchantable:
                    item.enchantment += 1
                    upgradedItemsCount += 1

                if item in player.gear.values():
                    item.equip()

        print("In a flash of cleansing light, all items in this floor have been uncursed.")
        if upgradedItemsCount > 0:
            print(f"{upgradedItemsCount} items have been upgraded.")

        super().consume(floor)

        self.degrade()
        return True

class ScrollEnchant(Scroll):
# adds 1 + INT/2 levels of enchantment to an item
    name = "scroll of enchantment"
    value = 65

    def inspect(self):
        print("Reading this scroll will bless one of your items.")
        print("It cannot be used on items with a higher level than your intelligence (INT).")
        super().inspect()

    def consume(self, floor):
        power = player.intelligence + self.enchantment
        
        # gathers input
        options = ["cancel"]
        for item in player.inventory:
            options.append(item.get_name())

        itemIndex = gather_input(f"Choose an item with a level of {c.green(power)} or less to bless.", options, True)

        if itemIndex == 0: # 0 cancels
            return False
        else:
            itemIndex -= 1 # converts to proper index
            chosenItem = player.inventory[itemIndex]

            if chosenItem.enchantment > power: # checks intelligence
                print(c.red(f"You can only upgrade items with a level of {power} or less."))
                return False
            
            if chosenItem == self:
                print(c.red("The scroll cannot bless itself!"))
                return False

            if chosenItem.enchantable: # checks if item is valid
                if chosenItem in player.gear.values():
                    chosenItem.unequip()

                chosenItem.enchantment += 1
                print(chosenItem.name + " has been blessed.")
                super().consume(floor)
                self.degrade()

                if chosenItem in player.gear.values():
                    chosenItem.equip()
                return True
            else:
                print(c.red(chosenItem.name + " cannot be enchanted."))
                return False

class ScrollRepair(Scroll):
# fully repairs and item, higher levels of int increase its max uses
    name = "scroll of repair"
    value = 50

    def inspect(self):
        print("This scroll will fully restore the uses of one item.")
        print("Your intelligence (INT) can increase or decrease the chosen item's future durability.")
        super().inspect()

    def consume(self, floor):
        power = player.intelligence + self.enchantment # intelligence boosts effectiveness

        # gathers input
        options = ["cancel"]
        for item in player.inventory:
            options.append(item.get_name())

        chosenItem = gather_input("What item do you repair?", options, True)

        if chosenItem == 0: # 0 cancels
            return False
        else:
            chosenItem -= 1
            item = player.inventory[chosenItem]
            if item.maxUses > 1 or item.uses < 1: # only works on items that have multiple max uses or have no uses remaining

                if item.maxUses > 1: # only improves items with multiple uses
                    item.maxUses += int(item.maxUses * power / 10)

                    if power > 0:
                        print(item.name + " is more durable now")
                    elif power < 0:
                        print(item.name + " is less durable now")
                    
                item.uses = item.maxUses
                print(item.name + " has been fully repaired")
                super().consume(floor)
                self.degrade()
                return True
            else:
                print(item.name + " cannot be repaired")
                return False

class Consumable(Item):
# a one use item that varies slightly from the scroll
    maxUses = 1

    def degrade(self):
        player.inventory.remove(self)
        return True

    def status(self):
        return ""

class Bomb(Consumable):
    name = "bomb"
    value = 45

    enchantable = True

    canDig = True

    def inspect(self):
        print("Can destroy walls, possibly revealing secret rooms.")
        print("Can be used in combat to damage all enemies.")
        if self.enchantment > 0:
            print(c.green("This bomb does extra damage."))
        elif self.enchantment < 0:
            print(c.red("This bomb does less damage."))

    def attack(self, enemies):
        for enemy in enemies:
            damage = randint(13, 15) + self.enchantment * 2

            # does less damage against bosses (discourages spamming bombs)
            if issubclass(type(enemy), entities.Boss):
                damage = damage // 2

            # bombs are unaffected by crit chance
            crit = player.critChance
            player.critChance = 0
            
            damage = enemy.hurt(player, damage, 1, 0)

            # restores crit chance
            player.critChance = crit
            
            print(f"The bomb does {c.red(damage)} damage to {enemy.name}.")

        self.degrade()
        return True

class StunBomb(Consumable):
    name = "stun bomb"
    value = 35

    enchantable = True

    def inspect(self):
        print(f"Leaves enemies {c.effect(entities.Dazed)} and {c.effect(entities.Surprised)}, lowering their DEX and PER.")
        print(f"Allows you to escape from any enemy (except bosses), and stuns enemies for a couple turns.")
        if self.enchantment > 0:
            print(c.green(f"Dazed lasts longer."))
        elif self.enchantment <= 0:
            print(c.red(f"Does not inflict Dazed."))

    def attack(self, enemies):
        for enemy in enemies:
            enemy.affect(entities.Surprised(), 2)
            if self.enchantment >= 0:
                enemy.affect(entities.Dazed(True), 2 + self.enchantment)
                
        print(f"All enemies are {c.effect(entities.Surprised)} and {c.effect(entities.Dazed)}!")
        self.degrade()
        return True

class FireBomb(Consumable):
    name = "fire bomb"
    value = 55

    enchantable = True

    def inspect(Self):
        print("Damages all enemies and sets them on fire.")

    def attack(self, enemies):
        for enemy in enemies:
            damage = randint(13, 15) + self.enchantment

            # does less damage against bosses (discourages spamming bombs)
            if issubclass(type(enemy), entities.Boss):
                damage = damage // 2

            # bombs are unaffected by crit chance
            crit = player.critChance
            player.critChance = 0
            
            damage = enemy.hurt(player, damage, 2, 0)

            player.critChance = crit
            
            enemy.affect(entities.OnFire(), 5 + self.enchantment)
            print(f"The fire bomb does {c.red(damage)} damage to {enemy.name}, and sets them {c.effect(entities.OnFire)}.")

        self.degrade()
        return True

class StorageBook(Consumable):
# consuming grants +1 inventory size
# purchased from the golem in the crossroads
    name = "\"Forbidden Techniques: Storage\""
    value = 45

    usePrompt = "read"

    def inspect(self):
        print("Contains knowledge on how to manipulate the physical world.")
        print("Reading will increase your inventory size.")

    def consume(self, floor):
        player.inventorySize += 1
        print("Most of the knowledge contained in the book is incomparehensible to your mortal mind.")
        print("Shortly after being read, the book burns to ashes. You gain +1 inventory size.")
        self.degrade()
        return True
        
class ImmunityBook(Consumable):
# consuming grants +1 immunity
# found in the mines
    name = "\"Forbidden Techniques: Immunity\""
    value = 60

    usePrompt = "read"

    def inspect(self):
        print("Contains knowledge on how to escape the limits of mortality.")
        print("Reading will increase your resistance.")

    def consume(self, floor):
        player.resistance += 1
        print("Most of the knowledge contained in the book is incomparehensible to your mortal mind.")
        print("Shortly after being read, the book burns to ashes. You gain +1 resistance.")
        self.degrade()
        return True

class EvasionBook(Consumable):
# consuming grants +5% dodge chance
# purchased from the collector
    name = "\"Forbidden Techniques: Evasion\""
    value = 100

    usePrompt = "read"

    def inspect(self):
        print("Contains knowledge on how to harness your shadow form.")
        print("Reading will increase your dodge chance.")

    def consume(self, floor):
        player.dodgeChance += 5
        print("Most of the knowledge contained in the book is incomparehensible to your mortal mind.")
        print("Shortly after being read, the book burns to ashes. You gain +5% dodge chance.")
        self.degrade()
        return True

class VisionBook(Consumable):
# consuming grants +1 awareness
# found as an artifact
    name = "\"Forbidden Techniques: Vision\""
    value = 100

    usePrompt = "read"

    def inspect(self):
        print("Contains knowledge on sensing the presence of others shadow forms.")
        print("Reading will increase your awareness.")

    def consume(self, floor):
        player.awareness += 1
        print("Most of the knowledge contained in the book is incomparehensible to your mortal mind.")
        print("Shortly after being read, the book burns to ashes. You gain +1 awareness.")
        self.degrade()
        return True

class Pickaxe(Item):
# can destroy tiles and is reusable
    name = "pickaxe"
    value = 70
    maxUses = 3

    enchantable = True

    canDig = True

    def status(self):
        message = ""
        if self.uses == 1:
            message = "(damaged)"
        elif self.uses == 2:
            message = "(worn)"

        return message

    def degrade(self):
        player.intelligence += self.enchantment
        super().degrade()
        player.intelligence -= self.enchantment

    def inspect(self):
        print("Can destroy walls, has multiple uses.")
        if self.enchantment > 0:
            print(c.green("This pickaxe has more durability."))
        elif self.enchantment < 0:
            print(c.red("This pickaxe has less durability."))

class IronKey(Item):
    name = "iron key"
    value = 20
    maxUses = 1

    def status(self):
        return ""

    def inspect(self):
        print(f"This key can open an iron gate.")

class GoldKey(Item):
    name = "gold key"
    value = 50
    maxUses = 1

    def status(self):
        return ""

    def inspect(self):
        print(f"This key can open a gold chest.")

class GoldChunk(Item):
    name = "gold chunk"
    value = 50

    def status(self):
        return ""

    def inspect(self):
        print(f"This is a chunk of unrefined gold.")

# number 1 through 16 is chosen
# if standardLoot = [(Rations, 8), (Bandage, 14), (Bomb, 16)]
# there is a 8 in 16 chance for rations, 6 in 16 chance for bandage, and 2 in 16 for bomb
class KnowledgeBook(Item):
# improves one stat
    name = "book of knowledge"
    value = 70
    maxUses = 1

    usePrompt = "read"

    def status(self):
        return ""

    def inspect(self):
        print("Reading this will let you level up one stat.")

    def consume(self, floor):
        print(f"{player.baseSTR} STR | {player.baseCON} CON | {player.baseDEX} DEX | {player.basePER} PER | {player.baseINT} INT")
        
        options = ["cancel", "STR", "CON", "DEX", "PER", "INT"]
        chosenStat = gather_input("What stat do you improve?", options, True, False)

        if chosenStat == "cancel":
            return False

        if chosenStat == "STR":
            player.update_str(1)
            player.baseSTR += 1
            print("Your attacks are stronger, you can carry more items.")
        elif chosenStat == "CON":
            player.update_con(1)
            player.baseCON += 1
            print("Your health has been increased, diseases and injuries heal quicker.")
        elif chosenStat == "DEX":
            player.update_dex(1)
            player.baseDEX += 1
            print("You are stealthier, you are better at dodging.")
        elif chosenStat == "PER":
            player.update_per(1)
            player.basePER += 1
            print("You're more aware, your attacks are better timed.")
        elif chosenStat == "INT":
            player.update_int(1)
            player.baseINT += 1
            print("Your items break less, scrolls are stronger.")
        
        player.inventory.remove(self)
        return True

class SorcerersRock(Item):
# increases wand recharge
    name = "Sorcerers Rock"
    value = 100
    maxUses = 1

    def status(self):
        return ""

    def inspect(self):
        print("Your wands recharge an additional charge when traveling between floors.")
    
    def pickup(self):
        player.recharge += 1

    def discard(self):
        player.recharge -= 1

class SeeingOrb(Item):
# reveals the whole map, requires a scroll of repair to recharge
# holding increases perception by 1
    name = "Seeing Orb"
    value = 100
    maxUses = 1

    usePrompt = "gaze"

    def status(self):
        if self.uses == 0:
            return "(uncharged)"
        else:
            return ""

    def consume(self, floor):
        if self.uses == 0:
            print("The orb needs to be charged with a scroll of repair before using it again.")
            return False

        self.uses -= 1

        floor.map = floor.layout
        print("You gaze into the orb, the whole floor has been revealed.")
        
        return True

    def pickup(self):
        player.update_per(1)

    def discard(self):
        player.update_per(-1)

    def inspect(self):
        print("Looking into the orb will reveal the entire layout of the floor.")
        print("Once used, it requires a scroll of repair to recharge it.")
        print("Having this item increases your perception (PER) by 1.")

rareLoot = [ShadowCloak, InfernoRing, IllusionRing, SeeingOrb, EbonyDagger, FlamingMace, CursedSword, EnchantedSpear, VisionBook]

def gen_loot(quality):
    if len(rareLoot) == 0: # replaces artifacts with scrolls if there aren't enough
        return ScrollEnchant()

    item = choice(rareLoot)
    rareLoot.remove(item)
    item = item()

    if item.enchantable:
        item.enchantment += (quality + randint(0, 1)) // 3
    return item

def pick_item(itemPool, itemNum):
# used in functions for generating weapons, armor, and wands
    for item in itemPool:
        if itemNum <= item[1]:
            return item[0]
        else:
            itemNum -= item[1]
    # returns the last item if the number is too large
    return itemPool[-1][0]


# returns a random item (usually a consumable)
def gen_item(quality):
    itemPool = [(Rations, 5), (Bomb, 5), (Bandage, 3), (StunBomb, 2), (Pickaxe, 1), (FireBomb, 2), (ScrollRepair, 1), (ScrollEnchant, 1)]

    item = pick_item(itemPool, randint(1, 12) + quality)()

    if item.enchantable and randint(1, 4) == 1: # 25% chance that item is blessed
        item.enchantment += 1
    
    return item

previousCategory = ["misc"] # this is a list b/c variables outside of functions can't be changed
# returns a random item that's gear (weapon, armmor, etc.)
def gen_gear(quality):
    categories = ["brute", "stealth", "misc"]
    categories.remove(previousCategory[0])

    category = choice(categories)
    
    gearPools = {
        "brute":[(Sword, 2), (Mace, 2), (Spear, 2), (Armor, 3), (LeatherArmor, 1)],
        "stealth":[(Dagger, 6), (Cloak, 3), (LeatherArmor, 1)],
        "misc":[(Ring, 10)]
    }

    gear = pick_item(gearPools[category], randint(1, 10))

    # tracks what category was selected
    if gear in (Sword, Mace, Spear, Armor):
        previousCategory[0] = "brute"
    elif gear in (Dagger, Cloak):
        previousCategory[0] = "stealth"
    else:
        previousCategory[0] = "misc"
    
    # some gear can be made of different materials, others get upgraded by being blessedd
    if gear in (Sword, Mace, Spear, Armor, Dagger):
        level = (quality + 2) // 3
        if level > 3:
            level = 3
            
        gear = gear(level)
    else:
        gear = gear()
        gear.enchantment += quality // randint(3, 4)
    
    if gear.enchantable: # gear can be cursed or blessed
        gear.enchantment += randint(-1, 1)
    
    return gear

def gen_scroll(quality = 0):
    scrollPool = ((ScrollEnchant, 2), (ScrollCleanse, 1), (ScrollRepair, 1))
    scroll = pick_item(scrollPool, randint(1, 4))()

    if randint(2, 19) < quality: # 5% chance to be blessed for each point of quality over 2
        scroll.enchantment += 1
        
    return scroll



from random import randint, choice
import textdungeon.items as items
import textdungeon.color as color
import textdungeon.entities as entities
from textdungeon.extra import clear_console, gather_input, slowprint, separator, pause, print_messages

c = color

player = entities.player

movementControls = [["move"], ["↑ North", "→ East", "↓ South", "← West",], ["↑", "→", "↓", "←",]][gather_input("Which movement controls do you prefer?", ["sub-menu", "arrows & words", "arrows only"])]

floors = []

def unlock(key):
    keyIndex = -1
    for item in player.inventory:
        if type(item) == key:
            keyIndex = player.inventory.index(item)

    if keyIndex == -1:
        print(f"You do not have a {key.name}.")
        return False
    else:
        options = ["cancel", "use key"]

        playerInput = bool(gather_input(f"Do you use a {key.name}?", options))

        if playerInput:
            player.inventory.pop(keyIndex)
            return True

        return False

def sort_inventory():
    weapons = []
    wands = []
    healing = []
    consumables = []
    apparel = []
    scrolls = []
    misc = []

    for item in player.inventory:
        if issubclass(type(item), items.Weapon):
            weapons.append(item)
        elif issubclass(type(item), items.Wand):
            wands.append(item)
        elif issubclass(type(item), items.Medicine):
            healing.append(item)
        elif issubclass(type(item), items.Scroll):
            scrolls.append(item)
        elif issubclass(type(item), items.Equipable):
            apparel.append(item)
        elif issubclass(type(item), items.Consumable):
            consumables.append(item)
        else:
            misc.append(item)

    player.inventory = weapons + wands + healing + consumables + apparel + scrolls + misc

def item_list():
# returns a list of item names
    itemList = []
    for item in player.inventory:
        itemList.append(item.get_name())

    return itemList

def update_effects(creature, enemies = None):
# iterates through every effect
    effectDurations = []
    
    for i in range(len(creature.effects)):
        creature.effects[i].update(enemies)
        creature.effects[i].duration -= 1

        effectDurations.append(creature.effects[i].duration)

    # deletes expired effects
    while 0 in effectDurations:
        effectIndex = effectDurations.index(0)
        effectDurations.pop(effectIndex)
        creature.effects[effectIndex].reverse()
        creature.effects.pop(effectIndex)

def print_effects(creature):
    effects = []
    for i in range(len(creature.effects)):
        effect = creature.effects[i]
        if effect.isPermanent:
            effects.append(f"{effect.color(effect.name)}")
        else:
            effects.append(f"{effect.color(effect.name)} - {creature.effects[i].duration} turns")

    if len(effects) > 0:
        print(f"[{' | '.join(effects)}]")

def print_player_info():
# prints player health, effects, stats, and equipment
    separator()
    print(f"YOU: [{c.health_status(player.health, player.maxHealth)} ♥] [{player.armorClass} AC]")
    
    print_effects(player)

    statsDisplay = ""

    statsDisplay += f"[{c.compare(player.strength, player.baseSTR)} STR] "
    statsDisplay += f"[{c.compare(player.constitution, player.baseCON)} CON] "
    statsDisplay += f"[{c.compare(player.dexterity, player.baseDEX)} DEX] "
    statsDisplay += f"[{c.compare(player.perception, player.basePER)} PER] "
    statsDisplay += f"[{c.compare(player.intelligence, player.baseINT)} INT] "
    
    print(statsDisplay)
    separator()
                        
class Battle:
    def __init__(self, enemies):
        self.battleOver = False
        self.enemies = enemies
        self.runChance = 100
        self.canRun = True

        self.allSurprised = False

    def start_battle(self):
        initialEnemies = len(self.enemies)
        
        
        clear_console()

        if player.has_effect(entities.Surprised) and not self.enemies[0].has_effect(entities.Surprised):
                slowprint(c.red("You are surprised!"))
        elif not player.has_effect(entities.Surprised) and self.enemies[0].has_effect(entities.Surprised):
                slowprint("Surprise attack!")
                if isinstance(type(self.enemies[0]), entities.Boss) and self.enemies[0].battleMessages != []:
                    slowprint(c.blue(choice(self.enemies[0].battleMessages)))
        elif self.enemies[0].battleMessages == []:
            slowprint(c.blue(f"You encounter {self.enemies[0].name}!"))
        else:
            slowprint(c.blue(choice(self.enemies[0].battleMessages)))

        if player.has_effect(entities.Surprised) and self.enemies[0].has_effect(entities.Surprised):
                slowprint("You surprise eachother.")
                self.allSurprised = True

        if self.enemies[0].isSpecial:
            slowprint(c.red("There is no escape from this fight."))

        pause()
        
        while not self.battleOver: 
            separator()
            # calculates run chance based on amount of enemies
            self.runChance = 100 + player.dodgeChance
            if len(self.enemies) > initialEnemies:
               self.runChance -= 70
            else:
                self.runChance -= int(70 / initialEnemies) * len(self.enemies)

            # checks each enemy to change run chance and check if player can run
            self.canRun = True
            for enemy in self.enemies:
                if enemy.stunned:
                    self.runChance += int(70 / initialEnemies)
                else:
                    self.runChance -= int(enemy.dodgeChance / initialEnemies)
                    
                if enemy.isSpecial or issubclass(type(enemy), entities.Boss):
                    self.canRun = False

            if self.canRun == False:
                self.runChance = 0

            if self.runChance > 100:
                self.runChance = 100
            elif self.runChance < 0:
                self.runChance = 0
            
            self.player_turn()
            
            if self.battleOver:
                break

            for enemy in self.enemies:
                if enemy.health > 0:
                    self.enemy_turn(enemy)
                    print_messages()

            # checks if enemies should die
            removedEnemies = []
            for enemy in self.enemies:
                if enemy.health <= 0:
                    print(f"{enemy.name} drops {enemy.gold} gold.")
                    player.gold += enemy.gold
                    removedEnemies.append(enemy)

            for enemy in removedEnemies:
                self.enemies.remove(enemy)

            if self.enemies == [] or player.health <= 0:
                self.battleOver = True
                break
            
        # detects if player won or lost
        if player.health > 0:
            return True
        else:
            return False

    def print_battle(self):
        for creature in self.enemies:
            print()
            print((c.red("(!) ") * creature.isSpecial) + f"{c.red(creature.name)} : [{c.health_status(creature.health, creature.maxHealth)} ♥] [{creature.armorClass} AC]")
            
            print_effects(creature)

        print_player_info()
    
    def enemy_turn(self, enemy):
        if self.allSurprised or not enemy.has_effect(entities.Surprised):
            enemy.do_turn(self.enemies)
        update_effects(enemy, self.enemies)

    def player_turn(self):
        turnOver = False

        if self.allSurprised or not player.has_effect(entities.Surprised):
            while not turnOver:
                self.print_battle()
    
                itemUsed = gather_input("What do you use?", [f"run ({self.runChance}% chance)"] + item_list(), True) - 1
    
                if itemUsed == -1:
                    if randint(0, 99) < self.runChance + 5: # a little bit of trickery
                        print(c.green("You succesfully escaped."))
                        self.run()
                    else:
                        print(c.red("You failed to escape."))
                    turnOver = True
                else:
                    turnOver = player.inventory[itemUsed].attack(self.enemies)
                    print_messages()
                    print()

        update_effects(player, self.enemies)

    def run(self): # called if the player chooses to run
        self.battleOver = True
        # heals the enemies a bit to make running less powerful
        for enemy in self.enemies:
            enemy.heal(5)
            enemy.awareness += 1
            enemy.stealth -= 1
            update_effects(enemy, self.enemies)

class Floor:
    def __init__(self, layout, posY, posX, entryMessage = ""): # posY and posX are player's position
        self.layout = layout
        self.posY = posY
        self.posX = posX
        self.entryMessage = entryMessage

        # forms the map
        self.map = []
        for i in range(len(self.layout)):
            self.map.append(['?'] * len(self.layout))

    # checks if an x or y coordinate is too big or small
    def check_pos(self, position):
        if position >= len(self.layout) or position < 0:
            return False
        else:
            return True

    # prints map
    def print_map(self):
        # prepares a list with the proper characters
        lines = []
        for i in range(len(self.map)):
            lines.append([])
            for I in range(len(self.map)):
                # player is represented as 'o'
                if i == self.posY and I == self.posX:
                    lines[i].append(c.blue('o'))
                # unknown is represented as '?'
                elif self.map[i][I] == '?':
                    lines[i].append('?')
                # walls are represented as '■'
                elif type(self.map[i][I]) == Wall and self.map[i][I].blocked:
                    if len(self.map[i][I].loot) > 0:
                        # walls with hidden loot are yellow
                        lines[i].append(c.yellow('■'))
                    else:
                        lines[i].append('■')
                # stairs are represented as ↓ or ↑
                elif "descend" in self.map[i][I].specialAction:
                    lines[i].append('↓')
                # shops and chests are represented as yellow !
                elif self.map[i][I].specialAction in ["shop", "unlock chest", "refine gold chunk"]:
                    lines[i].append(c.yellow('!'))
                # enemies are represented as red !
                elif self.map[i][I].check_detection():
                    lines[i].append(c.red('!'))
                # locked rooms appear as locks
                elif type(self.map[i][I]) == LockedRoom and self.map[i][I].blocked:
                    lines[i].append('x')
                # rooms are represented as ' '
                else:
                    lines[i].append(' ')
        # prints the map
        separator()
        for line in lines:
            print(" ".join(line))

    def update_room(self, y, x):
        self.map[y][x] = self.layout[y][x]
    # adds a room to the map
        
    def update_map(self):
    # adds nearby rooms to the map
    # warns player of nearby enemies
        y = self.posY
        x = self.posX

        messages = []
        nearbyEnemies = []
        
        if self.check_pos(y + 1):
            self.update_room(y + 1, x)
            nearbyEnemies.extend(self.layout[y + 1][x].threats)
            
        if self.check_pos(y - 1):
            self.update_room(y - 1, x)
            nearbyEnemies.extend(self.layout[y - 1][x].threats)
            
        if self.check_pos(x + 1):
            self.update_room(y, x + 1)
            nearbyEnemies.extend(self.layout[y][x + 1].threats)
            
        if self.check_pos(x - 1):
            self.update_room(y, x - 1)
            nearbyEnemies.extend(self.layout[y][x - 1].threats)
        
        for enemy in nearbyEnemies:
            if player.awareness >= enemy.stealth:
                if not enemy.warning in messages:
                    messages.append(enemy.warning)
                    print(enemy.warning)

    def move_player(self, direction):
    # moves the player
    # directions : 0 = up/north, 1 = right/east, 2 = down/south, 3 = left/west
        newY = self.posY + [-1, 0, 1, 0][direction]
        newX = self.posX + [0, 1, 0, -1][direction]

        # checks if new position is too big or small
        if not (self.check_pos(newY) and self.check_pos(newX)):
            print("The wall here is protected by magic and is indestructible.")
            return False

        # checks if new tile is a wall
        if self.layout[newY][newX].blocked:
            if self.layout[newY][newX].unblock():
                self.layout[newY][newX].blocked = False
            else:
                return False
        # updates players position
        self.posY = newY
        self.posX = newX
        return True

    def get_room(self):
    # retunrs the room that the player is in
        return self.layout[self.posY][self.posX]

    def get_options(self):
    # returns a list of options for the player to choose
        room = self.get_room()

        options = movementControls + ["wait", "view stats"]

        if player.inventory != []:
            options.append(f"inventory [{len(player.inventory)}/{player.inventorySize}]")
            
        if room.loot != []:
            options.append(c.yellow("take item"))
            # prints items
            if len(room.loot) == 1:
                print(f"There is a {c.yellow(room.loot[0].get_name())} here.")
            else:
                # gets a list of item names
                names = []
                for item in room.loot:
                    names.append(c.yellow(item.get_name()))
                    
                print(f"There is a {', '.join(names[0:-1])}, and a {names[-1]} here.")            

        # presents options to player and gathers input
        if room.specialAction != "":
            options.append(c.yellow(room.specialAction))

        if room.threats != []:
            options.append(c.red("surprise attack"))
            
            for enemy in room.threats:
                print(choice(enemy.stealthMessages))

            print(c.green("They do not notice you."))
    
        return options

    def action_move(self, direction = None):
    # called when player decides to move
        # - INPUT -
            playerInput = 0
            if direction == None:
                options = ["cancel", "↑", "→", "↓", "←"]

                self.print_map()
                separator()
                playerInput = gather_input("What direction do you move?", options, True) - 1

                if playerInput == -1: # -1 is cancel
                    return

            else:
                playerInput = ["↑", "→", "↓", "←"].index(direction) # delete if change is reverted

            # - MOVE -
            self.move_player(playerInput)
            
            self.update_map()
            update_effects(player)
    
            # - COMBAT -
            room = self.get_room()
            if room.threats != []:
                isNoticed = False
                isSurprised = True
                for enemy in room.threats: # checks if player is noticed
                    if enemy.awareness >= player.stealth:
                        isNoticed = True
                    if enemy.stealth <= player.awareness:
                        isSurprised = False
                
                if isNoticed: # is detected
                    if isSurprised: # is detected and suprised
                        player.affect(entities.Surprised(), 1)
                    battle = Battle(room.threats)
                    battle.start_battle()
                elif isSurprised: # is not detected but is surprised
                    player.affect(entities.Surprised(), 1)
                    for enemy in room.threats:
                        enemy.affect(entities.Surprised(), 1)
                    battle = Battle(room.threats)
                    battle.start_battle()
                elif issubclass(type(room.threats[0]), entities.Boss): # boss fight
                    for enemy in room.threats: # is not detected
                        enemy.affect(entities.Surprised(), 1)
                    battle = Battle(room.threats)
                    battle.start_battle()

    def action_wait(self):
    # called when player chooses to wait
        self.update_map()
        update_effects(player)
        print(choice(["You wait.", "You take a moment to enjoy your surroundings.", "You ponder existence.", "You fall asleep for an unknown amount of time."]))
        
    def action_view_stats(self):
    # called when player decides to view stats
        message = ""
        # strength
        message += f"{c.compare(player.strength, player.baseSTR)} STR : "
        message += f"you do {c.compare(player.bonusDamage, player.strength * 0.75)} extra damage with weapons "
        message += f"and you can carry up to {c.compare(player.inventorySize, player.strength + 10)} items"

        # constitution
        message += f"\n\n{c.compare(player.constitution, player.baseCON)} CON : "
        message += f"your maximum health is {c.compare(player.maxHealth, 20 + player.constitution * 2)}, "
        message += f"and your resistance level is {c.compare(player.resistance, player.constitution)}\n\t"

        minor = [] # minor resistances
        major = [] # major resistances

        for effect in [entities.Poisoned, entities.Burned, entities.Bleeding, entities.Chilled, entities.Dazed, entities.OnFire, entities.BrokenBones]:
            if effect.level < player.resistance:
                if player.resistance - effect.level > 3:
                    major.append(effect.name)
                else:
                    minor.append(effect.name)

        for effect in player.immuneTo:
            if not effect.name in major:
                major.append(effect.name)

        if len(minor) > 0:
            message += f"you have resistance to ({', '.join(minor)}) "
        else:
            message += "you have resistances "

        if len(major) > 0:
            message += f"and immunity to ({', '.join(major)})"
        else:
            message += "and no immunities"
        
        message += f"\n\n{c.compare(player.dexterity, player.baseDEX)} DEX : "
        message += f"you have a {c.compare(player.dodgeChance, player.dexterity * 5)}% chance to dodge attacks, and your stealth is level {c.compare(player.stealth, player.dexterity)}"
        
        message += f"\n\n{c.compare(player.perception, player.basePER)} PER : "
        message += f"your awareness is level {c.compare(player.awareness, player.perception)}, and you have a {c.compare(player.critChance, player.perception * 10)}% chance to deal double damage"
        
        message += f"\n\n{c.compare(player.intelligence, player.baseINT)} INT : "
        message += f"your gear degrades {100 - player.intelligence * 10}% of the time"

        print(message)

        print(f"\nYou have {player.armorClass} AC, granting {player.armorClass / 2} damage resistance.")

        # prints effects
        for i in range(len(player.effects)):
            effect = player.effects[i]

            title = effect.color(effect.name.upper())
            if not effect.isPermanent:
                title += f" ({player.effects[i].duration} turns remaining)"
            print()
            print(title)

            print(effect.inspect())

    def action_take_item(self):
    # called when player takes an item
        room = self.get_room()
        
        if len(player.inventory) >= player.inventorySize: # checks if inventory has space
            print(f"You are carrying too much, you can only have {player.inventorySize} items.")
            return

        # decides options
        options = ["cancel"]
        for item in room.loot:
            options.append(item.get_name())

        # gathers input if more than one item
        chosenItem = 0
        if len(room.loot) > 1:
            chosenItem = gather_input("What do you pickup?", options, True) - 1

        if chosenItem > -1: # -1 is cancel
            # moves item to inventory
            room.loot[chosenItem].pickup()
            player.inventory.append(room.loot.pop(chosenItem))
            print(f"You take the {options[chosenItem + 1]}.")
    
            sort_inventory()

    def action_inventory(self):
        while True:
            options = ["back"] + item_list()
            
            if player.gold > 0:
                print(f"You have {c.yellow(str(player.gold))} gold.\n")
            playerInput = gather_input("Select an item to inspect or use:", options, True) - 1

            # exits the loop if player selects "back"
            if playerInput == -1:
                break
            
            # figures out what item was selected and prepares options
            options = ["back"]
            chosenItem = player.inventory[playerInput]

            if chosenItem in player.gear.values():
                options.append("unequip")
                
            else:
                options.append("drop")

                if chosenItem.usePrompt != None:
                    options.append(chosenItem.usePrompt)

            # prints info about the item
            print(chosenItem.get_name())
            chosenItem.inspect()
            if chosenItem.enchantment > 0:
                print(f"\nThis item is {c.green('blessed')}.")
            elif chosenItem.enchantment < 0:
                print(f"\nThis item is {c.red('cursed')}.")

            # asks for input
            playerInput = gather_input("\nWhat do you do with " + chosenItem.get_name() + "?", options, True, False)

            if playerInput == chosenItem.usePrompt:
                chosenItem.consume(self)
                print_messages()

            elif playerInput == "drop":
                chosenItem.discard()
                self.get_room().loot.append(chosenItem)
                player.inventory.remove(chosenItem)

            elif playerInput == "unequip":
                chosenItem.unequip()

                for key in player.gear.keys():
                    if chosenItem == player.gear[key]:
                        player.gear[key] = None
                        break

                print("You unequip the " + chosenItem.get_name() + ".")

    def action_unlock_chest(self):
    # called when player uses an item
        room = self.get_room()
        
        room.unlock_chest()

    def action_surprise(self):
        room = self.get_room()
        
        if room.areEnemiesAware:
            print("You have already fought these enemies, they will not be surprised.")
            playerInput = gather_input("Are you sure you want to surprise attack?", ["cancel", "surprise attack"], True)

            if playerInput == 0:
                return

        else:
            for enemy in room.threats: # makes sure all enemies are surprised
                enemy.affect(entities.Surprised(), 1)

        battle = Battle(room.threats)
        battle.start_battle()

        room.areEnemiesAware = True
        
    def enter_floor(self):
    # starts when player enters the floor, ends when they exit
    # returns -1 or 1, representing the change in floor
        clear_console()

        if self.entryMessage != "":
            print(self.entryMessage)
        self.update_map()
        
        while player.health > 0:
            room = self.get_room()

            # ===== PRINTS INFORMATION =====
            self.print_map()
        
            if room.description != "":
                print(room.description)
            options = self.get_options()
            
            print_player_info()
            playerInput = gather_input("What do you do?", options, False, False)

            actions = {
                "move":self.action_move, "wait":self.action_wait,
                "view stats":self.action_view_stats, f"inventory [{len(player.inventory)}/{player.inventorySize}]":self.action_inventory,
                c.yellow("take item"):self.action_take_item,
                c.yellow("unlock chest"):self.action_unlock_chest, c.red("surprise attack"):self.action_surprise
            }
            
            if playerInput in actions.keys(): # some actions aren't in the dictionary
                actions[playerInput]()

            elif playerInput[0] in ["↑", "→", "↓", "←"]:
                self.action_move(playerInput[0])
                
            elif playerInput == "debug : reveal map":
                self.map = self.layout

            elif "descend" in playerInput:
                if room.interact(): # if it leads somewhere else it is in .interact()
                    break

            else:
                room.interact()

class Room:
    blocked = False # determines if it counts as a wall or not
    description = ""
    specialAction = ""

    areEnemiesAware = False # once a combat is over, this is set to True, these enemies cannot be surprised if it's True
    
    def __init__(self, loot = [], threats = [], description = ""):
        self.loot = loot
        self.threats = threats
        if description != "":
            self.description = description

    def unblock(self): # called if the room needs a bomb or key to unlock
        return True
    
    def check_detection(self): # checks if player notices enemies here
        detected = False
        for enemy in self.threats:
            if player.awareness >= enemy.stealth:
                detected = True

        return detected

    def interact(self): # called when player uses special action
        return True

class Refinery(Room):
# used to refine gold chunks
    description = "There is an old, Dwarven refinery here. It looks rather simple, and you think that you could operate it."
    specialAction = "refine gold chunk"

    def __init__(self):
        self.loot = []
        self.threats = []

    def interact(self):
        goldIndex = -1
        for i in range(len(player.inventory)):
            if type(player.inventory[i]) == items.GoldChunk:
                goldIndex = i
                break

        if goldIndex == -1:
            print(c.red("You have nothing to refine!"))
        else:
            player.inventory.pop(goldIndex)
            gold = randint(20, 50)
            player.gold += gold
            print(f"The refinery produces {c.yellow(gold)} gold from your gold chunk.")
            
class Chest(Room):
    blocked = False
    description = "There is a " + c.yellow("chest") + " here with a " + c.yellow("gold lock") + "."
    specialAction = "unlock chest"

    def __init__(self, depth):
        self.loot = []
        self.threats = []
        self.hiddenLoot = [items.gen_loot(depth)]

    def unlock_chest(self):
        if unlock(items.GoldKey):
            self.loot.extend(self.hiddenLoot)

            print("You unlock the chest.")
            self.description = ""
            self.specialAction = ""
            return True

        else:
            return False

class Wall(Room):
    blocked = True
    description = "You are in a tunnel, there is rubble everywhere."
    specialAction = ""

    def __init__(self):
        self.loot = []
        self.threats = []

    def unblock(self): # requires a bomb or pickaxe
        print("There is a wall in the way.")

        miningItems = []
        for item in player.inventory:
            if item.canDig:
                miningItems.append(item)
        
        if len(miningItems) > 0:
            # gathers input
            options = ["cancel"]
            for item in miningItems:
                options.append(item.get_name())
            
            itemUsed = gather_input("How do you destroy it?", options, True) - 1

            if itemUsed == -1:
                return False
            else:
                player.inventory[player.inventory.index(miningItems[itemUsed])].degrade()
                return True
        else:
            print(c.red("You have no way of destroying it.\n"))
            pause()
            return False

class LockedRoom(Room):
    blocked = True
    description = ""
    specialAction = ""

    def __init__(self, depth):
        self.loot = []
        self.threats = []

        self.loot = [items.gen_gear(depth + 1), items.gen_scroll(depth), choice([items.gen_item(depth + 2)])]

    def unblock(self): # requires a certain key
        print(f"This room is locked and requires a iron key.")

        if unlock(items.IronKey):
            print("The gate opens.")
            return True
        else:
            return False

class Chasm(Room):
    description = "There is a large chasm here."
    specialAction = "descend chasm"

    def interact(self):

        risk = c.red("major")
        if player.dexterity + player.constitution > 3:
            risk = c.green("minor")
        elif player.dexterity + player.constitution > 1:
            risk = c.yellow("moderate")

        prompt = f"Descending poses {risk} risk of injury, do you want to descend?"

        if gather_input(prompt, ["cancel", "descend"], True) == 0:
            return False
        
        g = Generator()
        g.initialize_floor("mines", 5, 7)

        outcome = 0
        if risk == c.green("minor"):
            outcome = randint(0, 1)
        elif risk == c.yellow("moderate"):
            outcome = randint(0, 2)
        else:
            outcome = randint(1, 2)

        if outcome == 0:
            g.entryMessage = c.blue("You safely descend the chasm.\n")
        elif outcome == 1:
            g.entryMessage = c.red("While climbing down the chasm you slip and sustain mild injury.\n")
            player.health -= randint(2, 5)
            player.affect(entities.Bleeding(), 3)
        else:
            g.entryMessage = c.red("While trying to climb down the chasm, you slip and break your bones.\n")
            player.health -= randint(4, 6)
            player.affect(entities.BrokenBones())
            player.affect(entities.Bleeding(), 4)
        
        g.generate_mines()
        floors.insert(1, g.finalize_floor())

        return True

class Stairs(Room):
    blocked = False
    description = "There are " + c.yellow("stairs") + " here that lead down."
    specialAction = "descend stairs"

    def __init__(self):
        self.loot = []
        self.threats = []

class Shop(Room):
    blocked = False
    description = "You stumble upon the " + c.yellow("SHOPKEEPER") + ", an ancient stone golem."
    specialAction = "shop"

    def __init__(self, depth):
        self.loot = []
        self.threats = []

        self.stock = []

        # has an enchanted gear of a next-area material
        self.stock.append(items.gen_gear(depth + 3))
        if self.stock[0].enchantable:
            self.stock[0].enchantment = randint(1, 2)
        
        # has 2 items or gear that can't be cursed
        for i in range(2):
            self.stock.append(items.gen_gear(depth))
            if self.stock[-1].enchantment < 0:
                self.stock[-1].enchantment = 0

        # has a scroll
        self.stock.append(items.gen_scroll(depth + 6))
        
        # has an item from the next area
        self.stock.append(items.gen_item(depth + 4))

        # has a healing item
        self.stock.append(choice([items.Bandage, items.Rations])())
        
        if depth == 4:
            self.stock.append(items.StorageBook())
        
        # items are more expensive in later floors
        for item in self.stock:
            item.value = int(item.value * (1.3 ** (depth // 3)))

    def interact(self):
        while True:
            options = ["cancel"]

            for item in self.stock:
                options.append(item.get_name() + f", costs {c.yellow(item.get_price())} gold")

            print(c.blue("The golem shows you it's wares."))
            playerInput = gather_input(f"You have {c.yellow(player.gold)} gold.", options, True) - 1

            if playerInput == -1: # cancel
                break
            else:
                purchase = self.stock[playerInput]

                if player.gold < purchase.get_price():
                    print(c.red(f"You need {purchase.get_price() - player.gold} more gold to purchase {purchase.get_name()}!"))

                elif len(player.inventory) == player.inventorySize:
                    print(c.red("Your inventory is full!"))
                    break

                else:
                    self.stock.remove(purchase)
                    player.gold -= purchase.get_price()
                    player.inventory.append(purchase)
                    sort_inventory()
                    print(f"You purchase {purchase.get_name()} for {purchase.get_price()} gold.")
                    break

class Collector(Room):
    blocked = False
    description = "You stumble upon " + c.purple("THE COLLECTOR") + ", an ancient stone golem."
    specialAction = "talk"

    def __init__(self):
        self.loot = []
        self.threats = []

        self.stock = [items.EvasionBook(), items.HealingVial(), items.gen_loot(8)]

        # sells some items from the next area
        self.stock.append(items.gen_item(8))
        self.stock.append(items.gen_item(8))

        # sells a guaranteed healing item
        self.stock.append(choice([items.Bandage, items.Rations])())
        
        # items are more expensive when bought from the collector
        for item in self.stock:
            item.value = int(item.value * 1.5)

    def interact(self):
        while True: # --- TODO: FINISH THIS SECTION ---
            options = ["cancel", "shop", "sell"]
            prompt = c.purple('"' + choice([
                "You don't want to know what's down there. Just give up now, and I'll gladly buy your items!",
                "Anything you don't sell to me now can be taken off of your dead body after you face my champion.",
                "Even though you survived the goblins, my champion will make short work of you."
                ]) + '"')
            playerInput = gather_input(prompt, options, True, False)

            if playerInput == "shop":
                while True:
                    options = ["cancel"]
        
                    for item in self.stock:
                        options.append(item.get_name() + f", costs {c.yellow(item.get_price())} gold")
        
                    print(c.purple("\"What do you wish to buy?\""))
                    playerInput = gather_input(f"You have {c.yellow(player.gold)} gold.", options, True) - 1
        
                    if playerInput == -1: # cancel
                        break
                    else:
                        purchase = self.stock[playerInput]
        
                        if player.gold < purchase.get_price():
                            print(c.red(f"You need {purchase.get_price() - player.gold} more gold to purchase {purchase.get_name()}!"))
        
                        elif len(player.inventory) == player.inventorySize:
                            print(c.red("Your inventory is full!"))
                            break
        
                        else:
                            self.stock.remove(purchase)
                            player.gold -= purchase.get_price()
                            player.inventory.append(purchase)
                            sort_inventory()
                            print(f"You purchase {purchase.get_name()} for {purchase.get_price()} gold.")
                            break
            elif playerInput == "sell":
                while True:
                    options = ["cancel"]

                    for item in player.inventory:
                        options.append(item.get_name() + " " + c.yellow(item.get_price() // 2))

                    playerInput = gather_input("Which item do you want to sell?", options, True) - 1

                    if playerInput < 0:
                        break
                    else:
                        item = player.inventory[playerInput]
                        print(f"Are you sure you want to {c.red('sell')} {c.yellow(item.get_name())} for {item.get_price() // 2} gold?")
                        playerInput = input("(y/n): ").lower()
                        clear_console()
                        if playerInput == "y":
                            print(f"You sold {c.yellow(item.get_name())} for {c.green(item.get_price() // 2)} gold.")
                            player.inventory.remove(item)
                            item.discard()
                            if item in player.gear.values():
                                item.unequip()

                            player.gold += item.get_price() // 2
            elif playerInput == "cancel":
                return

def gen_room(area, depth, type):
    loot = []
    threats = []
    room = Room()

    # standard room
    if type == 1:
        pass
        #if area == "prison":
         #   if randint(1, 3) == 1:
          #      room.description = "You are in an empty prison cell."

    # special room
    elif type == 2:
        pass
        #if area == "prison":
         #   roomDesc = randint(1, 6)
          #  if roomDesc == 1:
           #     room.description = "You are in an old storage room."
            #elif roomDesc == 2:
             #   room.description = "You are in what used to be an armory."
            #elif roomDesc == 3:
             #   room.description = "You are in a large, empty room."

    # secret room
    elif type == 3:
        room.description = "This is a secret room."

        if area == "mines":
            loot.append(items.ImmunityBook())
        else:
            loot.append(items.gen_loot(depth))
        
        for i in range(randint(1, 2)):
            loot.append(items.gen_item(depth + 3 - i))

    room.threats = threats
    room.loot = loot
    
    return room
    
class Generator:
# generates floors
    def initialize_floor(self, area, depth, size):
        self.area = area
        self.depth = depth
        self.size = size

        self.modifier = "" 
        if self.size > 4 and randint(0, 1): # can assign a random modifier except on the first floor
            self.modifier = choice(["dangerous", "large", "cursed"])
        self.entryMessage = ""

        # adds these features before finishing generation
        self.addRooms = []
        self.addItems = []
        self.addEnemies = []

        # selects the generation format early, so it can be overriden
        self.generation = choice([self.gen_hall, self.gen_intersection, self.gen_square])

    def generate_floor(self):
        # stores the layout
        self.layoutNums = [] # this is the layout in integers
        self.layoutRooms = [] # layoutNums is converted into rooms
        self.startY = 0
        self.startX = 0

        # stores info for room placement
        self.rooms = []
        self.adjacentWalls = []
        self.hiddenWalls = []
        self.sideRooms = []

        if self.modifier != "":
            self.entryMessage += {
                "flooded":c.blue("The ground is flooded with water."),
                "large":c.blue("Your footsteps echo across the floor."),
                "dangerous":c.red("This floor is unusually crowded, watch your back."),
                "cursed":c.red("A malevolent energy lurks in the items here.")
            }[self.modifier] + "\n"

        # applies "large" modifier
        if self.modifier == "large":
            self.size += 1

        # forms a square in layoutNums
        for i in range(self.size):
            self.layoutNums.append([0] * self.size)

        # mutates rats
        if self.area == "crossroads" and self.depth > 3:
            # checks which mutations can be made
            mutationList = ["toxic", "stronger", "hungrier"]
            for mutation in entities.Rat.mutations:
                mutationList.remove(mutation)
            
            mutation = choice(mutationList)

            entities.Rat.mutations.append(mutation)
            self.entryMessage += c.red({
                "toxic":"The rats have mutated.\n",
                "stronger":"The rats grow stronger.\n",
                "hungrier":"The rats are hungry.\n"
            }[mutation])

        self.generation()

        # generates secret room
        if len(self.hiddenWalls) > 0:
            hiddenSpot = choice(self.hiddenWalls)
            self.layoutNums[hiddenSpot[0]][hiddenSpot[1]] = 3

        # creates layoutRooms based on the numbers in layoutNums
        for y in range(self.size):
            self.layoutRooms.append([])
            for x in range(self.size):
                room = Wall() # 0 = Wall
    
                if self.layoutNums[y][x] > 0:
                    room = gen_room(self.area, self.depth, self.layoutNums[y][x])
                    
                elif self.layoutNums[y][x] == -1:
                    room = Stairs()
                
                self.layoutRooms[y].append(room)

        self.addItems.extend(self.gen_random_items(self.size + randint(0, 1), self.size - randint(0, 1)))
        
        # spawns enemies
        if self.modifier == "dangerous":
            self.addEnemies.extend(entities.gen_enemies(self.area, self.size, self.depth % 3, self.depth % 3))
        elif self.modifier == "flooded":
            self.addEnemies.extend(entities.gen_enemies(self.area, self.size - 2, self.depth % 3, self.depth % 3))
            choice(self.addEnemies).append(entities.SewerRat())
            self.addEnemies.append([entities.SewerRat(), entities.SewerRat()])
        else:
            self.addEnemies.extend(entities.gen_enemies(self.area, self.size - 1, self.depth % 3, self.depth % 3))

    def generate_mines(self):
        self.modifier = ""
        self.generation = self.gen_mine
        self.entryMessage += c.blue("The walls are rough and lined with gold, the air is still and filled with silence. You have entered The Mines.")
        
        self.generate_floor()
        self.layoutRooms[self.size // 2][0] = Refinery()

        self.addEnemies = [[entities.AncientDraugr()]] # only an ancient draugr
        self.addItems = [items.Pickaxe(), items.Bomb()] # no loot
        self.addRooms = [] # no side rooms

        walls = self.hiddenWalls + self.adjacentWalls
        for i in range(10):
            wall = choice(walls)
            walls.remove(wall)
            self.layoutRooms[wall[0]][wall[1]].loot.append(items.GoldChunk())

        # spawns worms
        walls = self.hiddenWalls + self.adjacentWalls
        # doesn't spawn worms near spawn
        walls.remove([0, 2])
        walls.remove([0, 4])
        for i in range(15):
            wall = choice(walls)
            if len(self.layoutRooms[wall[0]][wall[1]].threats) > 0:
                walls.remove(wall)
            self.layoutRooms[wall[0]][wall[1]].threats.append(entities.Worm())

    def gen_mine(self):
    # generates two intersecting halls in the middle, without generating side rooms
        for i in range(self.size): 
            self.layoutNums[i][self.size // 2] = 1
            self.layoutNums[self.size // 2][i] = 1

        self.layoutNums[self.size - 1][self.size // 2] = -1
        self.startX = self.size // 2
        self.startY = 0
        self.count_rooms()
           
    def gen_hall(self):
    # generates a snake-like hall
        genOver = False
        x, y = 0, 0
    
        self.layoutNums[x][y] = 1
    
        while not genOver:
            chosenDirection = choice(['x', 'y'])
    
            # keeps the directions somewhat even
            if (x - 2) > y:
                chosenDirection = 'y'
            elif (y - 2) > x:
                chosenDirection = 'x'
    
            if chosenDirection == 'x':
                # checks if x is too big
                if (x + 1) < self.size:
                    x += 1
                else:
                    break
    
            elif chosenDirection == 'y':
                # checks if y is too big
                if (y + 1) < self.size:
                    y += 1
                else:
                    break
    
            self.layoutNums[y][x] = 1
    
        # addds stairs down at the end
        self.layoutNums[y][x] = -1
        
        self.count_rooms()
        self.gen_rooms(((self.size * self.size) - len(self.rooms)) // 3)

    def gen_intersection(self):
    # generates two intersecting halls
        offset = 1 # how close to the edge the halls are allowed to be
        if self.size > 5:
            offset = 2

        hallX = randint(offset, self.size - 1 - offset)
        hallY = randint(offset, self.size - 1 - offset)

        self.layoutNums[hallX] = [1] * self.size

        for y in range(self.size):
            self.layoutNums[y][hallY] = 1

        self.startX = randint(0, self.size - 1)
        self.startY = hallX

        self.layoutNums[choice([0, self.size - 1])][hallY] = -1
        
        self.count_rooms()
        self.gen_rooms(((self.size * self.size) - len(self.rooms)) // 3)

    def gen_square(self):
    # generates a ring around that leaves one wall between it and the border
        self.layoutNums[1] = [0] + ([1] * (self.size - 2)) + [0]
        self.layoutNums[self.size - 2] = [0] + ([1] * (self.size - 2)) + [0]

        for y in range(1, self.size - 2):
            self.layoutNums[y][1] = 1
            self.layoutNums[y][self.size - 2] = 1

        self.startX = choice([1, self.size - 2])
        self.startY = choice([1, self.size - 2])

        endX = self.startX
        endY = self.startY

        while endX == self.startX and endY == self.startY:
            endX = choice([1, self.size - 2])
            endY = choice([1, self.size - 2])

        self.layoutNums[endY][endX] = -1
        
        self.count_rooms()
        self.gen_rooms(((self.size * self.size) - len(self.rooms)) // 3)

    def gen_random(self):
    # generates a random layout
        # picks starting position
        x = randint(1, self.size) - 1
        y = randint(1, self.size) - 1

        self.startX = x
        self.startY = y

        self.layoutNums[y][x] = 1

        # generation starts here
        previousDirection = -3

        for i in range(int(self.size * 1.5)):
            direction = randint(0, 3)
            stepSize = randint(2, 3)

            if direction == (previousDirection + 2) % 4:
                direction = (direction + 1) % 4

            for i in range(stepSize):
                if direction == 0 and y > 0:
                    y -= 1
                elif direction == 1 and x < self.size - 1:
                    x += 1
                elif direction == 2 and y < self.size - 1:
                    y += 1
                elif direction == 3 and x > 0:
                    x -= 1

                self.layoutNums[y][x] = 1

        while (x == self.startX and y == self.startY):
            direction = randint(0, 3)
            stepSize = randint(2, 3)

            for i in range(stepSize):
                if direction == 0 and y > 0:
                    y -= 1
                elif direction == 1 and x < self.size - 1:
                    x += 1
                elif direction == 2 and y < self.size - 1:
                    y += 1
                elif direction == 3 and x > 0:
                    x -= 1

                self.layoutNums[y][x] = 1

        self.layoutNums[y][x] = -1
        
        self.count_rooms()
        self.gen_rooms(((self.size * self.size) - len(self.rooms)) // 3)

    def count_rooms(self):
        for y in range(self.size):
            for x in range(self.size):
                if self.layoutNums[y][x] != 0:
                    self.rooms.append([y, x])
                else:
                    # checks for walls adjacent to rooms
                    adjacentToRoom = False
                    for i in [[y - 1, x], [y + 1, x], [y, x - 1], [y, x + 1]]:
                        # if coords are too large it passes
                        if (i[0] < self.size) and (i[1] < self.size) and (i[0] >= 0) and (i[1] >= 0):
                            # checks if viable for being an adjacent room 
                            if self.layoutNums[i[0]][i[1]] != 0:
                                self.adjacentWalls.append([y, x])
                                adjacentToRoom = True
                                break

                    # checks if viable for secret room
                    if not adjacentToRoom:
                        self.hiddenWalls.append([y, x])

    def gen_rooms(self, amount):
        viableLocations = [] + self.adjacentWalls
        for i in range(amount):
            if len(viableLocations) > 0:
                # picks a spot for a room
                location = choice(viableLocations)
                # places room
                self.layoutNums[location[0]][location[1]] = 2
    
                # changes lists accordingly
                self.rooms.append(location)
                self.sideRooms.append(location)
                self.adjacentWalls.remove(location)
                viableLocations.remove(location)

                y = location[0]
                x = location[1]

                # adds nearby exposed walls to adjacentWalls
                for i in [[y - 1, x], [y + 1, x], [y, x - 1], [y, x + 1]]:
                    if (i[0] < self.size) and (i[1] < self.size) and (i[0] >= 0) and (i[1] >= 0):
                        if self.layoutNums[i[0]][i[1]] == 0 and not i in self.adjacentWalls:
                            self.adjacentWalls.append(i)
                            self.hiddenWalls.remove(i)

    def gen_random_items(self, gearAmount, itemAmount):
        spawnedItems = []
        
        # spawns gear
        for i in range(gearAmount):
            randomItem = items.gen_gear(self.depth)

            # cursed modifier has a 1 in 3 chance to degrade every item
            if self.modifier == "cursed" and randomItem.enchantable and randint(1, 3) == 1:
                randomItem.enchantment -= 1

            spawnedItems.append(randomItem)

        # spawns items
        for i in range(itemAmount):
            randomItem = items.gen_item(self.depth)
            spawnedItems.append(randomItem)
        
        return spawnedItems

    def spawn_enemies(self):
        validRooms = self.rooms
        
        for enemy in self.addEnemies:
            # ERROR CATCHER
            if len(validRooms) == 0:
                for i in range(len(self.layoutNums)):
                    print(self.layoutNums[i])
                input("ERROR DETECTED, DISPLAYING MAP (ENTER TO DISMISS)")
            
            coords = choice(validRooms)
            room = self.layoutRooms[coords[0]][coords[1]]

            # spawns enemy
            room.threats.extend(enemy)

            validRooms.remove(coords)

    def spawn_rooms(self):
        for room in self.addRooms:
            # prefers side rooms
            if len(self.sideRooms) > 0:
                chosenSpot = choice(self.sideRooms)
                self.sideRooms.remove(chosenSpot)
                self.rooms.remove(chosenSpot)
                self.layoutRooms[chosenSpot[0]][chosenSpot[1]] = room
            else:
                chosenSpot = choice(self.rooms)
                self.rooms.remove(chosenSpot)
                self.layoutRooms[chosenSpot[0]][chosenSpot[1]] = room
    
    def spawn_items(self):
        validRooms = self.rooms + self.sideRooms # favors side rooms (they are included in rooms)
        
        for item in self.addItems:
            if len(validRooms) == 0:
                validRooms = self.rooms + self.sideRooms # resets valid rooms if there aren't enough

            coords = choice(validRooms)
            room = self.layoutRooms[coords[0]][coords[1]]

            # spawns enemy
            room.loot.append(item)

            validRooms.remove(coords)

    def finalize_floor(self):
        self.spawn_rooms()
        self.spawn_items()
        self.spawn_enemies()

        return Floor(self.layoutRooms, self.startY, self.startX, self.entryMessage)


from textdungeon.extra import add_message
from random import randint, choice
import textdungeon.color as color

c = color

class Creature:
    # identifying information (such as name or dialogue)
    name = "CREATURE"

    # the most basic stats
    maxHealth = 20
    gold = 0
    stealth = 0 # compared with opponents awareness in stealth checks
    awareness = 0 # compared with opponents stealth in awareness checks

    # other basic stats, not always modified by subclasses
    immuneTo = None # effects that cannot be applied
    armorClass = 0 # reduces incoming damage
    dodgeChance = 0 # percent chance to dodge
    resistance = 0 # higher values decrease the duration of certain effects
    critChance = 0 # percent chance to do x2 damage
    bonusDamage = 0 # damage added to spears, swords, and maces

    # these are mainly used for the player, mostly modify basic stats
    # note that besides strength and intelligence, changing these stats doesn't do much
    # use methods such as update_dex(num) instead of changing these stats directly
    strength = 0 # +damage +inventorySize
    dexterity = 0 # +dodge +stealth
    constitution = 0 # +maxHeatlh +resistance
    intelligence = 0 # increases gear durability and scroll effectiveness
    perception = 0 # +awareness +appraisal

    # temporary stats that keep track of combat status
    health = 0
    effects = None
    stunned = False
    canHeal = True

    def __init__(self):
        self.health = self.maxHealth

        self.immuneTo = []
        self.effects = []

    def update_str(self, increase):
    # strength is added to damage dealt and increases inventory size
        self.strength += increase
        self.bonusDamage += 0.75 * increase
        if issubclass(type(self), Player):
            self.inventorySize += increase

    def update_dex(self, increase):
    # dexterity improves stealth and dodge
        self.dexterity += increase

        self.stealth += increase
        self.dodgeChance += 5 * increase

    def update_con(self, increase):
    # constitution improves health and resistance
        self.constitution += increase

        self.resistance += increase
        self.maxHealth += 2 * increase
        self.health += 2 * increase

    def update_int(self, increase):
    # intelligence is how likely a used item or armor is to not break
        self.intelligence += increase
    
    def update_per(self, increase):
    # perception improves awareness and appraisal
        self.perception += increase

        self.awareness += increase
        self.critChance += 10 * increase
    
    def set_stats(self, str, con, dex, per, int):
    # sets all 5 stats at once
        self.update_str(str - self.strength)
        self.update_dex(dex - self.dexterity)
        self.update_con(con - self.constitution)
        self.update_int(int - self.intelligence)
        self.update_per(per - self.perception)

    def has_effect(self, effectType):
        for effect in self.effects:
            if type(effect) == effectType:
                return True
        return False

    def dodge(self, attacker):
        number = randint(0, 99)
        return self.dodgeChance > number

    def hurt(self, attacker, damage, piercing = 0, applyModifiers = True):
        bonusDamage = 0
        # apply modifiers enables crit chance and bonus damage
        if applyModifiers:
            bonusDamage += attacker.bonusDamage

            if attacker.critChance > 0:
                if randint(0, 100) < attacker.critChance:
                    bonusDamage += damage
                    add_message(attacker.name + " got a CRITICAL STIKE!")
            
        reduction = self.armorClass
        if reduction > 0 and piercing > 0: # piercing is only applied if AC > 0
            reduction -= piercing
            if reduction < 0: # piercing can't lead to -AC
                reduction = 0
                
        reduction /= 2.0 # each point of AC is only 0.5 damage reduction
        
        finalDamage = damage - reduction + bonusDamage
        
        # all attacks do at least 0.5 damage, (50% to do 1 damage)
        if finalDamage < 0.5:
            finalDamage = 0.5
        
        # if there is a decimal, it is treated as a chance to do extra damage
        # example: 5.4 damage has a 40% to do 6 damage, and 60% to do 5
        if (finalDamage != int(finalDamage)):
            dif = finalDamage - int(finalDamage)
            if randint(0, 99) < dif * 100:
                finalDamage += 1
        
        # calculations are made with floats but are returned as ints
        self.health -= int(finalDamage)
        return int(finalDamage)

    def heal(self, healthRestored):
    # heals health but makes sure it doesn't exceed max health
        if not self.canHeal:
            return 0
        
        finalHealthRestored = healthRestored
        if self.health + healthRestored > self.maxHealth:
            finalHealthRestored = self.maxHealth - self.health

        self.health += finalHealthRestored
        return finalHealthRestored

    def affect(self, effect, duration = 0):
    # applies resistance and immunities to an effect
        isPermanent = duration == 0
        
        # applies resistance
        if effect.natural and player.resistance > effect.level:
            if self.resistance - effect.level > 3:
                return False
            elif not isPermanent:
                duration -= self.resistance - effect.level
                if duration < 1:
                    return False

        # checks for duplicate effects
        for i in range(len(self.effects)):
            if type(effect) == type(self.effects[i]):
                # checks which effect is longer
                if self.effects[i].duration < duration or duration < 0:
                    self.effects[i].reverse()
                    self.effects.pop(i)
                    break
                else:
                    return False

        # checks if immune to effect
        if not type(effect) in self.immuneTo:
            if isPermanent:
                effect.isPermanent = True
                duration = -1
            else:
                effect.duration = duration
                
            self.effects.append(effect)
            effect.apply(self)
            return True
        else:
            return False

class Player(Creature):
# has inventory and equipment slots
    name = "YOU"
    maxHealth = 20

    gold = 40

    inventorySize = 10
    inventory = []
    gear = {"armor":None, "ring":None, "back":None}
    recharge = 1 # wand recharge speed

    baseSTR = 0
    baseCON = 0
    baseDEX = 0
    basePER = 0
    baseINT = 0

    currentFloor = None

    # various stats for unusual effects
    infernoRing = False
    illusionRing = False

    def set_stats(self, str, con, dex, per, int):
        super().set_stats(str, con, dex, per, int)
        self.baseSTR = str
        self.baseCON = con
        self.baseDEX = dex
        self.basePER = per
        self.baseINT = int

    def dodge(self, attacker):
        dodged = super().dodge(attacker)

        if dodged and self.illusionRing:
            attacker.stunned = True
            add_message(f"Your Ring of Illusion stuns {attacker.name}.")

        return dodged

    def hurt(self, attacker, damage, piercing = 0, strength = None):
    # damages armor
        damageDealt = super().hurt(attacker, damage, piercing, strength)

        if self.gear["armor"] != None:
            self.gear["armor"].degrade()

        # applies inferno ring's effect
        if self.infernoRing:
            self.affect(Burned(), 3 + (player.resistance // 2)) # duration of burned scales with resistance, making resistance less effective

        return damageDealt
    
    def affect(self, effect, duration = 0):
        isAffected = super().affect(effect, duration)

        if not isAffected:
            add_message("You resisted the effect!")
        elif type(effect) != Surprised: # surprised triggers too often and doesn't appear properly
            add_message(effect.color(effect.name.upper()) + ": " + effect.inspect())

        return isAffected

player = Player() # creates the one and only instance of player

class Enemy(Creature):
# subclasses of Enemy require a method named attack()
    # more identifying information
    battleMessages = [] # names are only colored in stealth due to the need to alert the player
    stealthMessages = [c.red("ENEMY") + " not notice you"]
    warning = "You feel uneasy..."
    undead = False
    isSpecial = False # determines if the game should give a (!) with this enemy
    
    def __init__(self):
        super().__init__()
        self.gold = (self.gold * randint(8, 12)) // 10
        self.awareness += randint(-1, 1)
        self.stealth += randint(-1, 1)

    def do_turn(self, enemies):
        # parameter 'enemies' allows the method to see the whole battlefield
        if self.stunned:
            print(f"{self.name} is stunned and unable to fight")
            self.stunned = False
        else:
            self.attack(enemies)

class Boss(Enemy):
    isSpecial = True

    def do_turn(self, enemies):
    # less likely to be stunned
        if self.stunned:
            if randint(0, 1):
                print(f"{self.name} is stunned and unable to fight")
                self.stunned = False
            else:
                print(f"{self.name} resisted the stun")
                self.stunned = False
                self.attack(enemies)
        else:
            self.attack(enemies)

class Effect:
    natural = False
    level = 0
    isPermanent = False
    color = c.effect_neutral

    duration = 0 # only changed when effect is created
    
    def apply(self, target):
        self.target = target

    def update(self, enemies):
    # called every turn
        pass

    def reverse(self):
    # called when effect is removed
        pass

    def inspect(self):
    # tells the player what the effect does
        return("This is an effect.")

class Electrocuted(Effect):
# stuns the target every other turn
    name = "electrocuted"
    color = c.effect_bad

    def update(self, enemies):
        if self.duration % 2 == 0:
            self.target.stunned = True

    def inspect(self):
        return("Stuns the target every other turn.")

class Bleeding(Effect):
# does 1 damage per turn
    name = "bleeding"
    natural = True
    level = 1
    color = c.effect_bad

    def update(self, enemies):
        # lowers health by 1 every turn
        self.target.health -= 1

    def inspect(self):
        return(f"Does {c.red(1)} damage every turn.")

class Regeneration(Effect):
# heals 1 hp per turn
    name = "regeneration"
    color = c.effect_good

    def update(self, enemies):
        self.target.heal(1)

    def inspect(self):
        return(f"Heals {c.green(1)} health every turn.")

class WellFed(Effect):
# heals 2 health per turn
    name = "well fed"
    color = c.effect_good

    def update(self, enemies):
        self.target.heal(2)

    def inspect(self):
        return(f"Heals {c.green(2)} health every turn.")
        
class Dazed(Effect):
# lowers DEX
    name = "dazed"
    natural = True
    level = 2
    color = c.effect_bad

    def __init__(self, allowRun = False):
        self.allowRun = allowRun
    
    def apply(self, target):
        self.target = target

        self.target.update_dex(-1)
        self.target.update_per(-1)

        self.allowRun = self.allowRun and self.target.isSpecial and not issubclass(type(target), Boss)

        if self.allowRun:
            self.target.isSpecial = False
        

    def reverse(self):
        self.target.update_dex(1)
        self.target.update_per(1)

        if self.allowRun:
            self.target.isSpecial = True

    def inspect(self):
        return(f"Lowers DEX and PER by {c.red(1)}.")

class Surprised(Effect):
# lowers dodge chance
    name = "surprised"
    color = c.effect_bad

    def apply(self, target):
        self.target = target

        self.target.dodgeChance -= 20

    def reverse(self):
        self.target.dodgeChance += 20

    def inspect(self):
        return(f"Lowers dodge chance by {c.red(20)}%.")

class Decay(Effect):
# lowers CON, gets stronger over time
    name = "decay"
    color = c.effect_bad

    def apply(self, target):
        self.target = target

        self.decayLevel = 1
        self.turnsToProgress = 2

        self.name = "decay lvl " + str(self.decayLevel)

        self.target.update_con(-self.decayLevel)

    def update(self, enemies):
        self.turnsToProgress -= 1

        if self.turnsToProgress == 0:
            self.decayLevel += 1
            self.name = "decay lvl " + str(self.decayLevel)
            self.turnsToProgress = self.decayLevel + 1
            self.target.update_con(-1)

    def reverse(self):
        self.target.update_con(self.decayLevel)

    def inspect(self):
        return(f"Lowers CON by {c.red(self.decayLevel)}, becomes stronger over time.")

class BrokenBones(Effect):
# lowers DEX, STR, CON, permanently
# instantly kills skeletons
    name = "broken bones"
    natural = True
    level = 5
    color = c.effect_bad

    def apply(self, target):
        self.target = target

        if issubclass(type(self.target), Skeleton):
            self.target.health = 0
            add_message(self.target.name + " dies")

        self.target.update_dex(-1)
        self.target.update_str(-1)
        self.target.update_con(-1)
        
    def reverse(self):
        self.target.update_dex(1)
        self.target.update_str(1)
        self.target.update_con(1)

    def inspect(self):
        return(f"Lowers STR, CON, and DEX by {c.red(1)}.")

class Burned(Effect):
# lowers AC by 1
    name = "burned"
    natural = True
    level = 0
    color = c.effect_bad

    def apply(self, target):
        self.target = target

        self.target.armorClass -= 1

    def reverse(self):
        self.target.armorClass += 1

    def inspect(self):
        return(f"Lowers armor class by {c.red(1)}.")

class OnFire(Effect):
# does 2 damage per turn
# also applies burned
    name = "on fire"
    natural = True
    level = 4
    color = c.effect_bad

    def apply(self, target):
        self.target = target

    def update(self, enemies):
        self.target.affect(Burned(), 5)
        self.target.health -= 2

    def inspect(self):
        return(f"Does {c.red(2)} damage every turn, inflicts {c.effect(Burned)}.")

class Poisoned(Effect):
# does 1 damage per turn and lowers STR
    name = "poisoned"
    natural = True
    level = -1
    color = c.effect_bad

    def apply(self, target):
        self.target = target
        
        self.target.strength -= 1

    def update(self, enemies):
        self.target.health -= 1

    def reverse(self):
        self.target.strength += 1

    def inspect(self):
        return(f"Lowers STR by {c.red(1)}, does {c.red(1)} damage every turn.")

class Chilled(Effect):
# does 1 damage per turn and lowers DEX
    name = "chilled"
    natural = True
    level = 2
    color = c.effect_bad

    def apply(self, target):
        self.target = target
        
        self.target.update_dex(-1)

    def update(self, enemies):
        self.target.health -= 1

    def reverse(self):
        self.target.update_dex(1)

    def inspect(self):
        return(f"Lowers DEX by {c.red(1)}, does {c.red(1)} damage every turn.")

class HealingBlocked(Effect):
# prevents healing
    name = "healing blocked"
    color = c.effect_bad

    def apply(self, target):
        self.target = target
        
        self.target.canHeal = False

    def reverse(self):
        self.target.canHeal = True

    def inspect(self):
        return("Prevents healing.")

class Rage(Effect):
# +2 STR
    name = "rage"
    color = c.effect_good

    def apply(self, target):
        self.target = target

        self.target.update_str(2)

    def reverse(self):
        self.target.update_str(-2)

    def inspect(self):
        return(f"Increases STR by {c.green(2)}.")

class SteelFlesh(Effect):
# +2 CON
    name = "steel flesh"
    color = c.effect_good

    def apply(self, target):
        self.target = target

        self.target.update_con(2)

    def reverse(self):
        self.target.update_con(-2)

    def inspect(self):
        return(f"Increases CON by {c.green(2)}.")

class Invisibility(Effect):
# +2 DEX
    name = "invisibility"
    color = c.effect_good

    def apply(self, target):
        self.target = target

        self.target.update_dex(2)

    def reverse(self):
        self.target.update_dex(-2)

    def inspect(self):
        return(f"Increases DEX by {c.green(2)}.")

class Cloaked(Effect):
# hides targets name, +10 dodge chance
    name = "cloaked"
    color = c.effect_good

    def apply(self, target):
        self.target = target

        self.target.dodgeChance += 10

        # hides and stores name
        self.targetsName = self.target.name
        self.target.name = "CLOAKED"

    def reverse(self):
        self.target.dodgeChance -= 10

        self.target.name = self.targetsName

    def inspect(self):
        return(f"Increases dodge chance by {c.green(10)}%, hides the targets name.")

class RatDisease(Effect):
# has 4 stages, each inheriting the effects of the last:
# 1) nothing, 2) -1CON, 3) -2INT, 4) lose max health over time
    name = "rat disease lvl 1"
    color = c.effect_bad

    stage = 1
    progression = 14

    def update(self, enemies):
        self.progression -= 1

        if self.progression == 0:
            self.progression = 14 - self.stage
            if self.progression < 8:
                self.progression = 8
            self.stage += 1

            if self.stage == 2:
                self.target.update_str(-1)
                print(c.red("You feel weaker."))
                self.name = "rat disease lvl 2"

            elif self.stage == 3:
                self.target.update_int(-2)
                print(c.red("Your mind becomes clouded."))
                self.name = "rat disease lvl 3"

            else:
                self.target.maxHealth -= 2
                self.target.health -= 3
                self.name = "rat disease lvl 4"
                if self.stage == 4:
                    print(c.red("You begin to decay."))
                else:
                    print(c.red("You continue to decay."))

    def reverse(self):
        if self.stage > 1:
            self.target.update_stregth(1)
            print(c.green("Your strength returns."))

        if self.stage > 2:
            self.target.update_int(2)
            print(c.green("Your mind clears."))

        if self.stage > 3:
            self.target.maxHealth += self.stage - 3
            print(c.red("Your body heals, but some damage remains."))

    def inspect(self):
        if self.stage == 1:
            return(f"Has no current effect, becomes stronger over time.")
        elif self.stage == 2:
            return(f"Lowers STR by {c.red(1)}, becomes stronger over time.")
        elif self.stage == 3:
            return(f"Lowers STR by {c.red(1)} and INT by {c.red(2)}, becomes stronger over time.")

        else:
            return(f"Lowers STR by {c.red(1)} and INT by {c.red(2)}, permanently lowers your max health over time.")
            
class Draugr(Enemy):
# a rare enemy that can appear in earlier floors
# starts with armor but it degrades when hurt
# can inflict bleeding
    name = "DRAUGR"
    stealthMessages = [c.red("DRAUGR") + " is on the hunt for human.",
                      c.red("DRAUGR") + " is seeking."]
    undead = True
    isSpecial = True

    maxHealth = 18
    gold = 16
    awareness = 3
    stealth = 2

    critChance = 10
    resistance = 2
    armorClass = 3

    def hurt(self, attacker, damage, piercing = 0, strength = None):
        damageDealt = super().hurt(attacker, damage, piercing, strength)

        if randint(0, 1) or damageDealt > 4:
            self.resistance -= 1
            self.armorClass -= 1
            self.maxHealth -= 1
            self.health -= 1
            add_message("DRAUGR's armor degrades.")
        
        return damageDealt
        
    def attack(self, enemies):
        if player.dodge(self):
            print(f"You dodge DRAUGR's attack.")
            return

        if randint(1, 3) == 1:
            if player.affect(Bleeding(), 4):
                damage = player.hurt(self, 5)
                print(f"DRAUGR hits you with their axe for {c.red(damage)} damage, leaving you {c.effect(Bleeding)}!")
        else:
            damage = player.hurt(self, 6)
            print(f"DRAUGR hits you with their axe for {c.red(damage)} damage!")

class Ghoul(Enemy):
# an uncommon, more aware enemy that appears in the prison
# can dodge attacks and inflicts decay
    name = "GHOUL"
    warning = "You smell a foul stench..."
    stealthMessages = [c.red("GHOUL") + " is roaming.",
                      c.red("GHOUL") + " is waiting for human, they have yet to notice you."]
    undead = True

    maxHealth = 16
    gold = 10
    awareness = 4
    stealth = 1
    
    dodgeChance = 10
    strength = -1 # 4 damage felt like too much but 3 is too little, so this nerfs 4 a little bit

    def attack(self, enemies):
        if randint(1, 3) == 1:
            if player.affect(Decay(), 6):
                print(f"GHOUL curses you with {c.effect(Decay)}!")
                return
        
        if player.dodge(self):
            print("You dodge GHOUL's bite.")
            return

        damage = player.hurt(self, 4)
        print(f"GHOUL bites you for {c.red(damage)} damage!")

class Skeleton(Enemy):
# a common enemy type throughout the dungeon
# is immune to most natural effects and often staggers instead of attacking
    name = "SKELETON"
    warning = "You hear the shuffling of bones..."
    undead = True

    maxHealth = 15
    gold = 8
    awareness = 1
    stealth = 0
    
    damage = 3
    staggerChance = 2 # _ in 6
    
    def __init__(self):
        super().__init__()
        self.immuneTo.extend([Bleeding, Poisoned])

        if self.name == "SKELETON": # doesn't apply to subclasses
            self.weapon = choice(["sword", "spear", "mace"])
            self.battleMessages = [f"SKELETON grips their {self.weapon}!",
                                  "SKELETON finally gets to see some action!"]
            self.stealthMessages = [c.red("SKELETON") + f" is holding a {self.weapon}."]

    def do_turn(self, enemies):
    # there is a chance that skeletons stagger and don't attack
        if randint(0, 5) < self.staggerChance:
            print(f"{self.name} staggers and misses their attack.")
            self.stunned = False
        else:
            super().do_turn(enemies)

    def attack(self, enemies):
        # checks dodge
        if player.dodge(self):
            print(f"You dodge SKELETONS {self.weapon}.")
            return

        piercing = 0
        effect = None

        # spears can pierce
        if self.weapon == "spear" and randint(0, 1):
            piercing = 1

        # some weapons can inflict effects
        if randint(1, 4) == 1:
            # swords inflict bleeding
            if self.weapon == "sword":
                if player.affect(Bleeding(), 4):
                    effect = Bleeding

            # maces inflict dazed
            if self.weapon == "mace":
                if player.affect(Dazed(), 2):
                    effect = Dazed
        
        # does damage
        if effect == None:
            damage = player.hurt(self, self.damage, piercing)
            print(f"{self.name} attacks you with their {self.weapon} for {c.red(damage)} damage!")
        else:
            damage = player.hurt(self, self.damage, piercing)
            print(f"{self.name} attacks you with their {self.weapon} for {c.red(damage)} damage, leaving you {c.effect(effect)}!")

class SkeletonGuard(Skeleton):
# has more AC, staggers less, always has a spear, very aware
    name = "SKELETON GUARD"
    warning = "You hear the shuffling of bones..."
    stealthMessages = [c.red("SKELETON GUARD") + " had promised not to fall asleep this time."]
    isSpecial = True

    maxHealth = 17
    gold = 15
    awareness = 5
    stealth = -1

    armorClass = 4
    immuneTo = [Bleeding, Burned]

    staggerChance = 1 # _ in 6
    weapon = "spear"

class Thief(Enemy):
# an uncommon, stealthy and aware enemy
# hits you with a poison dart when at full health, might run away later in combat
    name = "THIEF"
    warning = "You are being watched..."
    battleMessages = ["\"I see you got some items there, the collector will be pleased.\"",
                     "You encounter a goblin THIEF!",
                     "THIEF equips a dart!"]
    stealthMessages = [c.red("THIEF") + " is looking for a victim.",
                      c.red("THIEF") + " is preparing poisons, and is unaware of your presence.",
                      "A goblin " + c.red("THIEF") + " is roaming."]

    maxHealth = 16
    gold = 6
    awareness = 2
    stealth = 4
    
    critChance = 5

    time = 0
    hasDart = True

    def do_turn(self, enemies):
    # as time goes on, they are more likely to run away
        super().do_turn(enemies)
        
        self.time += 1
        if randint(1, 4) < self.time:
            enemies.remove(self)
            print("THIEF runs away!")
            
        elif self.time == 3:
            print(choice(["THIEF seems eager to escape.", "THIEF wants to flee."]))

    def attack(self, enemies):
        if not self.hasDart:
            if player.dodge(self):
                print("You dodge THIEF's dagger.")
                return

            damage = player.hurt(self, 4)
            print(f"THIEF stabs you for {c.red(damage)} damage!")
        else:
            self.hasDart = False
            if player.dodge(self):
                print("You dodge THIEF's dart.")
                return
            
            if player.affect(Poisoned(), 6):
                print(f"THIEF hits you with a dart, inflicting {c.effect(Poisoned)}!")
            else:
                player.health -= 1
                print("THIEF hits you with a dart, but you resist its poison.")
    
class Ogre(Boss):
# big enemy, can inflict dazed, bleeding, and broken bones
    name = "OGRE"
    warning = "You hear the sounds of an ogre..."
    battleMessages = ["\"Long time it's been since human dared wander down here, you make tasty treat.\"",
                     "\"Me hungry, human tasty.\"",
                     "\"I CRUSH YOU!\""]

    maxHealth = 34
    gold = 40
    awareness = 3
    stealth = -1
    
    armorClass = 1
    resistance = 2
    dodgeChance = -10

    isRaged = False
    isCharging = False
    previousMove = "heavy"

    def attack(self, enemies):
        # becomes stronger when below 20 HP
        if self.health < 20 and not self.isRaged:
            self.isRaged = True
            self.bonusDamage += 2
            print("OGRE is enraged!")
        
        # choses move, can't do the same twice in a row
        choices = ["heavy", "slam", "attack"]
        choices.remove(self.previousMove)
        chosenMove = choice(choices)

        if not self.isCharging:
            self.previousMove = chosenMove

        if self.isCharging:
            self.isCharging = False

            player.dodgeChance += 10
            if player.dodge(self):
                print("You manage to dodge OGRE's heavy swing.")
                player.dodgeChance -= 10
                return
            player.dodgeChance -= 10
            
            damage = player.hurt(self, 7, 2)

            if damage > 8:
                if player.affect(BrokenBones()):
                    print(f"OGRE hits you with a heavy strike, dealing {c.red(damage)} and inflicting {c.effect(BrokenBones)}!")
                    return
            
            print(f"OGRE hits you with a heavy strike, dealing {c.red(damage)} damage!")
            
        elif chosenMove == "heavy":
            print("OGRE prepares a heavy swing!")
            self.isCharging = True

        elif chosenMove == "slam":
            if player.dodge(self):
                print("OGRE slams the ground, but you get out of the way.")
                return

            damage = player.hurt(self, 3, 3)
            if player.affect(Dazed(), 2):
                print(f"OGRE slams the ground, dealing {c.red(damage)} damage and leaving you {c.effect(Dazed)}!")
            else:
                print(f"OGRE slams the ground, dealing {c.red(damage)} damage!")            

        else:
            if player.dodge(self):
                print("You dodge OGRE's club!")
                return

            damage = player.hurt(self, 4, 2)

            if player.affect(Bleeding(), 3):
                print(f"OGRE hits you with their club, dealing {c.red(damage)} damage, leaving you {c.effect(Bleeding)}!")
            else:
                print(f"OGRE hits you with their club, delaing {c.red(damage)} damage!")

class Rat(Enemy):
# a weak enemy who spawns in large groups
# can inflict self with decay, then infects the player
    name = "RAT"
    warning = "You hear the rats..."
    battleMessages = ["You encounter a group of RATS!"]
    stealthMessages = ["You spot a decayed " + c.red("RAT") + ".",
                      c.red("RAT") + " is eating. They are a foul, decayed creature.",
                      "You see a hungry " + c.red("RAT") + ".",
                      c.red("RAT") + " is roaming."]

    maxHealth = 12
    gold = 5
    awareness = 2
    stealth = 1
    
    corruption = 0
    mutations = []
    isToxic = False
    isHungry = False

    def __init__(self):
        if "stronger" in self.mutations:
            self.maxHealth += 4
        
        self.isToxic = "toxic" in self.mutations
        self.isHungry = "hungrier" in self.mutations

        self.maxHealth -= randint(1, 4)

        super().__init__()

        self.health -= randint(0, 1)

    def do_turn(self, enemies):
        super().do_turn(enemies)

        if randint(0, 1):
            self.corruption += 1

    def attack(self, enemies):
        # inflicts self with decay
        if self.corruption == 1:
            effect = Decay
            if self.isToxic:
                effect = Poisoned

            self.affect(effect())
            print(f"RAT becomes infected with {c.effect(effect)}.")

        # rats are easier to dodge
        player.dodgeChance += 15
        if player.dodge(self):
            print("RAT leaps at you, but you avoid them.")
            player.dodgeChance -= 15
            return
        player.dodgeChance -= 15

        # inflicts player with decay
        if self.corruption == 1:
            self.corruption += 1
            
            bonusDuration = 0
            effect = Decay
            if self.isToxic:
                effect = Poisoned
                bonusDuration += 4

            # effect last longer if you already have it
            for i in range(len(player.effects)):
                if type(player.effects[i]) == effect:
                    bonusDuration = player.effects[i].duration - 1
                    break

            if player.affect(effect(), 4 + bonusDuration):
                damage = player.hurt(self, 4)
                print(f"RAT bites you for {c.red(damage)} damage, infecting you with {c.effect(effect)}!")
                return
            else:
                damage = player.hurt(self, 4)
                print(f"RAT bites you for {c.red(damage)} damage, but you resist {c.effect(effect)}!")
                return
        
        # eats teammate
        if self.health < 8 and randint(0, 1) and len(enemies) > 1 and self.isHungry:
            selfIndex = enemies.index(self)
            
            enemies.remove(self)
            target = choice(enemies)

            enemies.insert(selfIndex, self)

            damage = target.hurt(self, 3)
            healing = self.heal(5)
            print(f"RAT bites their teammate {target.name} for {c.red(damage)} damage, healing themselves {c.green(healing)} health!")
            return
            
        # nibbles through armor
        if randint(0, 2) == 2 and player.gear["armor"] != None:
            piercing = 2
            if player.armorClass > 4:
                piercing += 1
            player.gear["armor"].degrade()
            damage = player.hurt(self, 4, piercing)
            print(f"RAT nibbles through your armor, {c.red('degrading')} it and dealing {c.red(damage)} damage!")
            return

        # standard attack
        if self.isHungry and randint(0, 1):
            damage = player.hurt(self, 4)
            healing = self.heal(2)
            print(f"RAT bites you for {c.red(damage)} damage, restoring {c.green(healing)} health!")
        else:
            damage = player.hurt(self, 4)
            print(f"RAT leaps at you, biting you for {c.red(damage)} damage!")

class SewerRat(Enemy):
# a slightly tougher rat that can inflict bleeding and has an armor piercing attack
# if the player is bleeding, it can inflict rat disease
    name = "SEWER RAT"
    warning = "You hear the rats.."
    battleMessages = ["You encounter SEWER RAT!",
                      "SEWER RAT has found its next meal!"]
    stealthMessages = ["You find a mutated " + c.red("SEWER RAT") + ".",
                       "You spot an infected " + c.red("SEWER RAT") + ".",
                       c.red("SEWER RAT") + " is eating the decayed corpse of another rat."]
    
    maxHealth = 14
    gold = 5
    awareness = 4
    stealth = 1

    armorClass = 1
    dodgeChance = 5

    def attack(self, enemies):
        if player.dodge(self):
            print("You dodge SEWER RAT's attack.")
            return
        
        if randint(0, 1) and player.gear["armor"] != None:
            player.gear["armor"].degrade()
            damage = player.hurt(self, 3, 2)

            print(f"SEWER RAT nibbles through your armor, {c.red('degrading')} it and dealing {c.red(damage)} damage!")
            return
        
        else:
            damage = player.hurt(self, 3)

            isBleeding = player.has_effect(Bleeding)

            if isBleeding:
                player.affect(RatDisease())
                print(f"SEWER RAT bites you for {c.red(damage)} damage, infecting your wound with {c.effect(RatDisease)}!")
                return
            
            else:
                player.affect(Bleeding(), randint(4, 5))
                print(f"SEWER RAT bites you for {c.red(damage)} damage, leaving you {c.effect(Bleeding)}!")
                return


class RatBeast(Enemy):
# tough enemy, rages when low health, loses hp over time
# can bite and ram, starts injured and with a random effect
    name = "RAT BEAST"
    warning = "You hear a loud wheezing..."
    stealthMessages = [c.red("RAT BEAST") + " is wandering.",
                      c.red("RAT BEAST") + " is looking for their next meal.",
                      "You encounter " + c.red("RAT BEAST") + ", a rat the size of a bear."]

    maxHealth = 32
    gold = 14
    awareness = 5
    stealth = 1

    immunity = -1
    armorClass = 1

    isRaged = False

    def __init__(self):
        self.maxHealth -= randint(0, 2)
        super().__init__()
        self.health -= randint(0, 4)

        effect = choice([Bleeding, Decay, Dazed, Poisoned, Regeneration, WellFed, Burned])
        self.affect(effect(), randint(4, 6))

    def do_turn(self, enemies):
        super().do_turn(enemies)

        if randint(0, 1):
            print("RAT BEAST decays.")
            self.health -= randint(2, 3)
            self.maxHealth -= randint(1, 2)

        if self.health < 16 and not self.isRaged:
            self.isRaged = True
            self.strength += 2
            print(choice([
                "RAT BEAST is becoming desperate!",
                "RAT BEAST is enraged!"
            ]))

    def attack(self, enemies):
        if randint(0, 1):
            if player.dodge(self):
                print("RAT BEAST lunges at you, but you dodge.")
                return

            damage = player.hurt(self, 6)
            if randint(0, 1):
                player.affect(Bleeding(), randint(5, 7))
                print(f"RAT BEAST bites you for {c.red(damage)} damage, inflicting {c.effect(Bleeding)}!")
            else:
                print(f"RAT BEAST bites you for {c.red(damage)} damage!")

        else:
            player.dodgeChance += 5
            if player.dodge(self):
                print(f"You evade RAT BEAST and they hit a wall, leaving them {c.effect(Dazed)}.")
                self.affect(Dazed(), 2)
                player.dodgeChance -= 5
                return
            player.dodgeChance -= 5

            damage = player.hurt(self, 5, 3)
            player.affect(Dazed(), randint(1, 2))
            print(f"RAT BEAST rams you for {c.red(damage)} damage, leaving you {c.effect(Dazed)}!")

class Goblin(Enemy):
# dexterous enemy, has a dart that blocks healing and has a bandage
    name = "GOBLIN"
    warning = "You hear goblins..."
    battleMessages = ["You encounter a GOBLIN scout!",
                     "\"I'll be taking your gold!\"",
                     "\"You can't escape the collector!\""]
    stealthMessages = [c.red("GOBLIN") + " is waiting.",
                      c.red("GOBLIN") + " is keeping watch.",
                      c.red("GOBLIN") + " is sleeping."]

    maxHealth = 14
    gold = 9
    awareness = 4
    stealth = 4

    armorClass = 2
    critChance = 5

    hasDart = True
    hasBandage = True

    def attack(self, enemies):
        if self.hasBandage and self.health < 9:
            self.hasBandage = False
            healing = self.heal(4)
            self.affect(Regeneration(), 4)
            print(f"GOBLIN uses a bandage, restoring {c.green(healing)} health and applying {c.effect(Regeneration)}.")

        elif self.hasDart and player.health < 13 and not player.has_effect(HealingBlocked):
            self.hasDart = False
            if player.dodge(self):
                print("You dodge GOBLIN's dart.")
                return
            player.affect(HealingBlocked(), 3)
            print(f"GOBLIN hits you with a poisoned dart, inflicting you with {c.effect(HealingBlocked)}!")

        else:
            if player.dodge(self):
                print("You dodge GOBLIN's attack.")
                return
            damage = player.hurt(self, 4)
            print(f"GOBLIN attacks you for {c.red(damage)} damage!")

class BuffedGoblin(Goblin):
# same as goblin, but spawns alone and is stronger
    maxHealth = 16
    gold = 13

    armorClass = 3
    dodgeChance = 15
    strength = 2

class Trickster(Boss):
# can cloak self and allies, high dodge chance and attacks when they dodge list.insert(index, item)
    name = "TRICKSTER"
    battleMessages = ["\"I'm not allowed to let you pass, boss's orders.\"", "\"I'm the boss's champion, you're not going to get past me.\""]

    maxHealth = 30
    gold = 30
    
    armorClass = 2
    dodgeChance = 20
    
    hasSmoke = True
    
    def dodge(self, attacker):
        dodged = super().dodge(attacker)

        if dodged:
            player.affect(Poisoned(), 7)
            add_message(f"TRICKSTER retaliates by hitting you with dart, inflicting {c.effect(Poisoned)}!")

        return dodged
        
    def hurt(self, attacker, damage, piercing = 0, strength = None):
        damage = super().hurt(attacker, damage, piercing, strength)

        if self.health <= 0 and self.hasSmoke: # prevents being one-hit killed
            self.health = 1

        return damage

    def attack(self, enemies):
        if self.health < 10 and self.hasSmoke: # smoke bomb
            self.hasSmoke = False

            # summons 2 goblins and changes position
            enemies.extend([Goblin(), Goblin()])
            enemies.remove(self)
            enemies.insert(randint(0, 2), self)

            # affects all with cloaked and weakens the goblins
            for enemy in enemies:
                enemy.health = 18
                enemy.maxHealth = 18
                enemy.armorClass = 0
                
                for effect in enemy.effects:
                    effect.reverse()

                enemy.effects = []

                if enemy != self:
                    enemy.affect(Cloaked(), randint(4, 5))
                    enemy.strength -= 1
                    enemy.hasDart = False
                    enemy.hasBandage = False
                else:
                    self.isSpecial = False

            # affects self for longer because the effect triggers right after turn
            self.affect(Cloaked(), randint(5, 6))
            
            print(f"Two GOBLINs enter the arena, and TRICKSTER uses a smoke bomb, affecting all enemies with {c.effect(Cloaked)}!")
        else:
            damage = player.hurt(self, randint(4, 5), 2)
            print(f"TRICKSTER stabs you for {c.red(damage)} damage!")
     
class Alchemist(Enemy):
# buffs allies with Rage (+2 STR), Steel Flesh (+2 CON), and Invisibility (+2 DEX)
# runs away if alone
    name = "ALCHEMIST"
    warning = "You hear goblins..."
    battleMessages = ["You encounter ALCHEMIST!",
                      "\"The goblins just can't get enough of my potions!\"",
                     "\"The collector doesn't need to know about your death!\"",
                     "\"He wants to meet you, but I think I'll kill you instead!\""]
    stealthMessages = [c.red("ALCHEMIST") + " is mixing potions.",
                      c.red("ALCHEMIST") + " is selling potions to the goblins."]
    
    maxHealth = 20
    gold = 10
    stealth = 2
    awareness = 2

    armorClass = 1
    dodgeChance = 10

    def __init__(self):
        super().__init__()
        self.affect(Regeneration(), 5)

    def attack(self, enemies):
        if len(enemies) == 1:
            enemies.remove(self)
            print("ALCHEMIST runs away!")
        else:
            options = []
            for enemy in enemies:
                if enemy != self:
                    options.append(enemy)
            target = choice(options)

            potion = choice([Rage, SteelFlesh, Invisibility])
            target.affect(potion(), 5)
            
            print(f"ALCHEMIST gives {target.name} a potion of {c.effect(potion)}!")

class Worm(Enemy):
# appears in the walls of the mines
# can block healing and latch onto the player to drain their health
    name = "WORM"
    warning = c.red("You hear a soft rumbling in the walls...")
    battleMessages = ["You have disturbed WORM!",
                      "WORM has awoken!",
                      "You encounter WORM!"]
    stealthMessages = ["You see a sleeping " + c.red("WORM") +"."]

    maxHealth = 17
    gold = 3
    awareness = 7
    stealth = 1

    resistance = 3
    armorClass = 1

    isLatched = False

    def hurt(self, attacker, damage, piercing = 0, strength = None):
        damageDealt = super().hurt(attacker, damage, piercing, strength)

        if self.isLatched:
            player.affect(Bleeding(), 6)
            add_message(f"WORM detaches from you, leaving you {c.effect(Bleeding)}!")
            self.name = "WORM"
            self.isLatched = False
        
        return damageDealt

    def attack(self, enemies):
        if self.isLatched:
            damage = player.hurt(self, 4, 4)
            healing = self.heal(damage)
            print(f"WORM drains you for {c.red(damage)} damage, healing them for {c.green(healing)} health!")
        else:
            if player.dodge(self):
                print("You dodge WORM's attack!")
                return

            if randint(0, 1):
                damage = player.hurt(self, 5, 1)
                player.affect(HealingBlocked(), 4)
                print(f"WORM's venomous bite does {c.red(damage)} damage, and inflicts {c.effect(HealingBlocked)}!")
            else:
                self.isLatched = True
                self.name = "WORM (latched)"
                print(c.red("WORM latches onto you!"))

class AncientDraugr(Enemy):
# a rare enemy that can appear in the crossroads
# can throw axe, then switches to magic
    name = "ANCIENT DRAUGR"
    battleMessages = ["You encounter ANCIENT DRAUGR!",
                     "ANCIENT DRAUGR has killed many, and doesn't intend to stop!"]
    stealthMessages = [c.red("ANCIENT DRAUGR") + " is on the hunt for human.",
                        c.red("ANCIENT DRAUGR") + " has a freezing aura."]
    undead = True
    isSpecial = True

    maxHealth = 28
    gold = 26
    awareness = 8
    stealth = 2
    
    resistance = 2
    armorClass = 4

    hasAxe = True
        
    def attack(self, enemies):
        if self.hasAxe:
            if self.health < 26 and randint(0, 1):
                # axe throw
                self.hasAxe = False
                player.dodgeChance += 10
                if player.dodge(self):
                    print("ANCIENT DRAUGR throws their axe, but you dodge it.")
                    player.dodgeChance -= 10
                    return
                player.dodgeChance -= 10
                    
                damage = player.hurt(self, 6)
                player.affect(Bleeding(), 7)
                print(f"ANCIENT DRAUGR throws their axe at you, dealing {c.red(damage)} damage and inflicting {c.effect(Bleeding)}!")
            else:
                # standard attack
                if player.dodge(self):
                    print("You dodge ANCIENT DRAUGR's attack.")
                    return

                damage = player.hurt(self, 5, 1)
                print(f"ANCIENT DRAUGR attacks you with their axe for {c.red(damage)} damage!")
        else:
            # magic attack
            damage = player.hurt(self, randint(3, 4), 3)
            player.affect(Chilled(), 4)
            print(f"ANCIENT DRAUGR uses frost magic, dealing {damage} damage and inflicting {c.effect(Chilled)}!")

enemyPool = {
    "prison":[([Skeleton], 6), ([Thief], 3), ([Ghoul], 3)],
    "crossroads":[([Rat, Rat], 3), ([Rat, Rat, Rat], 3), ([RatBeast], 3), ([BuffedGoblin], 3)],
    "mines":[([Worm], 12)]
} # each number means _ in 12 chance
# enemies are ordered weakest to strongest

specialEnemyPool = {
    "prison":[([SkeletonGuard], 6), ([Skeleton, Skeleton], 3), ([Draugr], 3)],
    "crossroads":[([Alchemist, Goblin, Goblin], 8), ([AncientDraugr], 4)],
    "mines":[([AncientDraugr], 12)]
} # special enemies are stronger and less common

def gen_enemies(area, normalEnemies, specialEnemies, dangerModifier = 0):
    enemies = []

    while len(enemies) < normalEnemies:
        # enemies with higher numbers are stronger
        enemyNum = randint(1, 12) + dangerModifier

        # compares enemyNum to the number in each tuple
        for fight in enemyPool[area]:
            # checks if the number is small enough or it is the last enemy
            if enemyNum <= fight[1] or enemyPool[area].index(fight) == len(enemyPool[area]) - 1:
                addedEnemies = []
                for enemy in fight[0]:
                    addedEnemies.append(enemy())

                enemies.append(addedEnemies)

                break
            # prepares enemyNum for next iteration
            else:
                enemyNum -= fight[1]

    # same as above except with special enemies
    for i in range(specialEnemies):
        enemyNum = randint(1, 12)

        for fight in specialEnemyPool[area]:
            if enemyNum <= fight[1] or specialEnemyPool[area].index(fight) == len(specialEnemyPool[area]) - 1:
                addedEnemies = []
                for enemy in fight[0]:
                    addedEnemies.append(enemy())

                enemies.append(addedEnemies)

                break
            else:
                enemyNum -= fight[1]

    return enemies



import os
from time import sleep

"""
TABLE OF CONTENTS :

messages - used to store messages that'll be printed to the screen

add_message(message) - adds a message to messages

print_messages - prints and clears messages

clearCommand - tracks what command to give the system to clear the console

clear_console() - erases all text on the screen using a system command

gather_input(prompt, options, startAtZero, returnInt) - gathers input by presenting options and requesting a number

slowprint(text, speed = .03) - prints the text one character at a time, with speed seconds in-between characters

separator() - prints a long line of '-' to separate sections of text

pause() - prompts the user to press enter to continue, erases prompt but not previous text afterwards

"""

messages = [] # this is here because it needs to be accessed by all other files

# this exists b/c things in extra are imported separately and I don't want to have a var named messages in other files (too common)
def add_message(message):
    messages.append(message)

def print_messages():
    if len(messages) > 0:
        for message in messages:
            print(message)

        messages.clear()

        print()

# if on a slow machine it doesn't slow print
doSlowPrint = True
#if input("Optimize performance? (y/n): ").lower() == 'y':
 #   doSlowPrint = False
    
# detects the os and selects the appropriate clear command
clearCommand = 'clear'
if os.name in ('nt', 'dos'):
    clearCommand = 'cls'

def clear_console():
# used to delete all text on the screen
    os.system(clearCommand)

def gather_input(prompt, options, startAtZero = False, returnInt = True):
# INPUT : prompt (string), options (list), returnInt (bool, optional)
# OUTPUT : provides the player with options and requires an INT as a response, returns index or name of response
    # prints and numbers options
    print(prompt)
    for i in range(len(options)):
        print(f"[{i + int(not startAtZero)}] {options[i]}")

  # gathers input and accounts for user error
    validResponse = False
    while not validResponse:
        playerInput = None
        try:
            playerInput = input("Enter a Number : ")
            playerInput = int(playerInput) - int(not startAtZero)
        
            if playerInput in range(len(options)):
                validResponse = True
            elif playerInput < 0:
                print(f"'{playerInput + int(not startAtZero)}' is too small")
            else:
                print(f"'{playerInput + int(not startAtZero)}' is too large")
        except ValueError:
            print(f"'{playerInput}' is not a number")

    clear_console()

    if returnInt:
        return playerInput
    else:
        return options[playerInput]

# prints text slower
def slowprint(text, speed=.03):
    if doSlowPrint:
        for letter in text:
            print(letter, end="", flush=True)
            sleep(speed)
        print()
    else:
        print(text)

def pause():
    input("- PRESS ENTER -")
    print("\033[1A\033[K", end="")

# break between output
def separator(end="\n"):
    print("----------------------------------------", end=end)


# \033[1;32m

regular = "\033[0m"
"""
red = "\033[1;31m" # danger, enemies, bad, damage
yellow = "\033[1;33m" # loot, gold
green = "\033[1;92m" # good, healing
blue = "\033[1;36m" # descriptive, player
purple = "\033[1;35m" # story
"""

# used for enemies, to highlight damage, and for negative events, things, and effects
def red(text):
    return "\033[1;31m" + str(text) + regular

# used to highlight loot, loot providing encounters, or as a transition from green to red
def yellow(text):
    return "\033[1;33m" + str(text) + regular

# used for beneficial effects and to highlight healing or other positive things
def green(text):
    return "\033[1;92m" + str(text) + regular

# used to highlight speech, descriptive text, or mysterious encounters
def blue(text):
    return "\033[1;36m" + str(text) + regular

# used for gods and powerful effects
def purple(text):
    return "\033[1;35m" + str(text) + regular

# these functions are give to Effect classes and are called whenever they need to display their name
# effect colors has effect as an argument because python tries to input self into them
def effect_good(effect, text):
    return green(text)

def effect_neutral(effect, text):
    return yellow(text)

def effect_bad(effect, text):
    return red(text)

def effect_great(effect, text):
    return purple(text)

# compare is primarily used for displaying stats, the color determines if they're being modified or not
def compare(actual, base):
    if actual < base:
        return red(actual)
    elif actual > base:
        return green(actual)
    else:
        return str(actual)

# requires a subclass of Effect, displays the effect with it's appropriate color
def effect(eff):
    return eff.color(eff, eff.name.upper())


# used in health_status, determines the color assigned to each status of health
def critical_health(text):
    return red(text)

def low_health(text):
    return yellow(text)

def full_health(text):
    return green(text)

# displays the message of health/maxHealth with appropriate color
def health_status(health, maxHealth):
    text = f"{health}/{maxHealth}"
    if health / maxHealth < 0.40:
        return critical_health(text)
    elif health / maxHealth < 0.75:
        return low_health(text)
    else:
        return full_health(text)